#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include <pspkernel.h>
#include <pspthreadman.h>
#include <psprtc.h>
#include <pspgu.h>
#include <pspgum.h>
#include <pspdisplay.h>

#include "main.h"

int gameModeTimer;
enum GameMode gameMode;
enum GameMode newMode;
extern enum GameMode pauseMode;
extern enum GameMode newPauseMode;
ScePspFVector3 from={15,15,15},to={0,0,0},up={0,1,0};
ScePspFVector3 fromTarget={45,45,45};
ScePspFVector3 reticle={240,131,0};
#define MAXENEMYMODEL 32
struct AnimModel enemyModel[MAXENEMYMODEL],heroModel[5];
int enemyModelCount;

struct Teather {
	ScePspFVector3 from,to;
	struct Vertex3DCP *vert;
	int vertCount;
	int when;
};

enum BonusType { BT_NOTHING, BT_LEVELUP, BT_CHAOSUP, BT_EVOUP, BT_EVODOWN, BT_MUDKIP };

struct Stage {
	unsigned long trigger; // in seconds * 1000 since beginning of segment.
	ScePspFVector3 pos;
	ScePspFVector3 orient;
};

enum EnemyType { ET_HERO, ET_ACTIVE, ET_DECORATION, ET_EVOUP, ET_CHAOSUP, ET_LEVELUP};
#define MAXBONUSTRIGGER 64
struct Actor;
#define MAXTRIGGER 512
struct Trigger {
	char type[64];
	int segment;	// activation segment
	struct Stage stage[8];
	int stageCount;
	enum EnemyType enemyType;
	enum BonusType bonusType;
	int health;
	int points;

	struct Actor *actor;
} trigger[MAXTRIGGER],bonusTrigger[MAXBONUSTRIGGER];
int triggerCount;
int bonusTriggerCount;
int targeted;

#define MAXENEMY 512
struct Actor {
	int active;
	struct AnimModel *model;
	ScePspFVector3 pos;
	ScePspFVector3 orient;
	enum EnemyType enemyType;
	// MD2 State information
	enum MotionType activeMotion;		// frame animation
	enum MotionType requestedMotion;
	int lastFrame;
	int frame;
	int frameElapsed;	// time since beginning of activeAnim in sec*1000
	int frameDuration;	// total time for this frame in sec*1000
	// Trigger
	struct Trigger *trigger;	// Contains the path for the current trigger path, or null.
	int health;
	int targeted;
	struct Teather teather;
	float minMax[6];
} enemy[MAXENEMY],hero;
int enemyCount;
int enemyHit;
int enemyMiss;
int supportHit;
int supportMiss;
int gameScore;
int shownScore;
int evoUp;	// maxes at 8, but then does an upgrade.
int evoKills;	// must be 8 to get a evoUp
int chaosUp;	// maxes at 3
int chaosKills;	// must be 44 to get a chaosUp
int chaosLeft;
const int chaosMax=2000;	// timer length when chaos starts in sec*1000.
int lookAround;
int showScore=1;
int scoreBonus=0;
int showFps=0;
int invulnerableTimer=0;

struct Explosion {
	int active;	// flag of whether the explosion is active.
	ScePspFVector3 pos;
	ScePspFVector3 orient;
	int elapsed;	// time since beginning of activeAnim in sec*1000	
	int duration;
	unsigned long firstColor;
	unsigned long lastColor;
	struct Vertex3DCNP *vert;
	int vertCount;
	int stripCount;
} explosion[32];
const int explosionCount=32;

enum CameraAngle { CA_FORWARD, CA_LEFT, CA_RIGHT };
struct Segment {
	int isBoss;		// boss level?
	int bossHealth;	// simple count down.
	int isTunnel;	// tunnel or horizon
	enum CameraAngle cameraAngle;
	int visualization;
	Image *heightMap;
	Image *colorMap;
	float seed[2];	// for the visualization
	float length;
	float scale;	// altitude of the terrain
	u32 fogColor;
	float fogNear;
	float fogFar;
	u32 palette[256];	// 256 palette entries.
	int duration;	// in seconds*1000
	int elapsed;
};

#define MAXSEGMENT 16
struct Level {
	char name[128];
	struct Segment segment[MAXSEGMENT];
	int segmentCount;
	int duration;	// in seconds*1000
	int elapsed;	// in seconds*1000
	int activeSegment;	// last active segment along the path
	float position;	// percent/100 of way through path.
	float bpm;	// beats per minute
	int evoDropRate;
	int chaosDropRate;
	int initialEvo;
	int initialChaos;
	int bossHealth;
	int bossHealthMax;
	int bossTimer;
	char checkSum[256];
} level;

int currentDir=0;
struct Heading { int x,y; } dir[]={{0,-1},{1,0},{0,1},{-1,0}};
float px=0,py=0;	// push x,y
int lockOn=0;

Image *reticleImage[9];
Image *hudLeftImage;
Image *hudRightImage;
Image *hudBossImage;
Image *chaosImage;

void genTerrainNoPoly();
void updateCamera(unsigned long elapsed);
void triggerExplosion(struct Actor *enemy);
void dropMessage(enum BonusType bt,struct Actor *actor);

/** Project a point from object coordinates to screen coordinates. 
 * \param view view transformation matrix. 
 * \param projection projection transformation matrix (for example from gumPerspective). 
 * \param model model transformation matrix (for example from gumLookAt). 
 */ 
ScePspFMatrix4 view;
ScePspFMatrix4 fastProjectMatrix;
void fastPreProject(const ScePspFMatrix4 *view,const ScePspFMatrix4 *projection,const ScePspFMatrix4 *model)
{
	ScePspFMatrix4 m2;
	gumMultMatrix(&m2,view,projection);
	gumMultMatrix(&fastProjectMatrix,&m2,model);
}

/** Project a point from object coordinates to screen coordinates. Prep with fastPreProject.
 * \param object point in object coordinates. 
 * \param viewport the left, top, width and height of the viewport. 
 * \param screen on success, the result of the point projected into screen coordinates. 
 * \returns 1 on success, or 0 if it is a degenerate projection. 
 */ 
s32 fastProject(const ScePspFVector3 *object,const s32 viewport[4],ScePspFVector3 *screen)
{
	float w;
	screen->x=object->x*fastProjectMatrix.x.x+object->y*fastProjectMatrix.y.x+object->z*fastProjectMatrix.z.x+fastProjectMatrix.w.x;
	screen->y=object->x*fastProjectMatrix.x.y+object->y*fastProjectMatrix.y.y+object->z*fastProjectMatrix.z.y+fastProjectMatrix.w.y;
	screen->z=object->x*fastProjectMatrix.x.z+object->y*fastProjectMatrix.y.z+object->z*fastProjectMatrix.z.z+fastProjectMatrix.w.z;
	w=object->x*fastProjectMatrix.x.w+object->y*fastProjectMatrix.y.w+object->z*fastProjectMatrix.z.w+fastProjectMatrix.w.w;
	
	if(w==0) return 0;
	screen->x/=w;
	screen->y/=w;
	screen->z/=w;
	
	// Now map x,y to the range 0-1
	screen->x=screen->x*0.5f+0.5f;
	screen->y=screen->y*0.5f+0.5f;
	// Scale/translate to the viewport size
	screen->x=screen->x*viewport[2]+viewport[0]; // *width + left
	screen->y=(1.0f-screen->y)*viewport[3]+viewport[1]; // *height+top
	return 1;
} 

void loadAnimModel(struct AnimModel *model,const char *md2fname,const char *texfname)
{
	const char *s=md2fname;
	if(strrchr(s,'\\')) s=strrchr(s,'\\')+1;
	if(strrchr(s,'/')) s=strrchr(s,'/')+1;
	strcpy(model->name,s);
	model->cullBackface=1;
	model->mesh[0]=md2Load(md2fname);
	if(!model->mesh[0]) {
		printf("Could not find model '%s'\n",md2fname);
		return;
	}
	printf("Loaded mesh for model '%s'\n",md2fname);
	model->mesh[0]->texture=loadPng(texfname);
	swizzleFast(model->mesh[0]->texture);
	model->motion[MT_IDLE].startFrame=0;
	model->motion[MT_IDLE].loopFrame=0;
	model->motion[MT_IDLE].loopLength=30;
	if(model->mesh[0]->frameCount<30) {
		model->motion[MT_IDLE].loopLength=model->mesh[0]->frameCount;
	} else if(model->mesh[0]->frameCount>=198) {
		model->motion[MT_IDLE].startFrame=40;	// Running animation of stock model.
		model->motion[MT_IDLE].loopFrame=40;
		model->motion[MT_IDLE].loopLength=6;
	}
	model->motion[MT_HIT].startFrame=0;
	model->motion[MT_HIT].loopFrame=0;
	model->motion[MT_HIT].loopLength=0;
}

int findEnemyModel(const char *fname)
{
	const char *base=strrchr(fname,'/');
	if(!base) base=strrchr(fname,'\\');
	if(!base) base=fname; else base++;
	int i;
	for(i=0;i<enemyModelCount;i++) {
		if(strcasecmp(base,enemyModel[i].name)==0) {
			return i;
		}
	}
	if(enemyModelCount>=MAXENEMYMODEL) {
		printf("Out of enemyModel slots for '%s'. Using default model.\n",base);
		for(i=0;i<enemyModelCount;i++) {
			printf("Model %d: '%s'\n",i,enemyModel[i].name);
		}
		return 0;	// could not load!
	}
	char path[256];
	strcpy(path,"models/");
	strcat(path,fname);
	char imageName[256];
	strcpy(imageName,path);
	if(strrchr(imageName,'.')) strrchr(imageName,'.')[0]=0;
	strcat(imageName,".png");
	loadAnimModel(enemyModel+i,path,imageName);
	if(!enemyModel[i].mesh[0]) {
		printf("Model load failed for '%s'\n",fname);
		return 0;
	}
	sceKernelDcacheWritebackAll();	// for the PNG.
	enemyModelCount++;
	return i;
}

void initActor(struct Actor *actor,struct AnimModel *model)
{
	int i;

	actor->active=1;
	actor->enemyType=ET_ACTIVE;
	actor->model=model;
	// MD2 state
	actor->frame=0;
	actor->lastFrame=0;
	actor->activeMotion=MT_INACTIVE;
	actor->requestedMotion=MT_INACTIVE;
	actor->frameElapsed=0;
	actor->frameDuration=1000/30;
	// rest
	actor->enemyType=ET_ACTIVE;
	actor->health=1;
	actor->orient.x=0; actor->orient.y=0; actor->orient.z=0;
	actor->pos.x=0; actor->pos.y=0; actor->pos.z=0;
	actor->targeted=0;
	actor->teather.vert=0;
	actor->teather.vertCount=0;
	actor->teather.when=0;
	actor->trigger=0;
	for(i=0;i<6;i++) {
		actor->minMax[i]=0;
	}
}

void initExplosion(struct Explosion *explosion,ScePspFVector3 *pos,ScePspFVector3 *orient)
{
	explosion->active=1;
	explosion->duration=1000;	// 1 second
	explosion->elapsed=0;
	explosion->pos.x=pos->x;
	explosion->pos.y=pos->y;
	explosion->pos.z=pos->z;
	explosion->orient.x=orient->x;
	explosion->orient.y=orient->y;
	explosion->orient.z=orient->z;
	explosion->firstColor=GU_RGBA(48,0,0,255);
	explosion->lastColor=GU_RGBA(48,48,0,255);
	if(explosion->vert) {
		free(explosion->vert);
		explosion->vert=0;
	}
}

void newLevel()
{
	strcpy(level.name,"level001.txt");
	level.duration=3*60*1000;
	level.segmentCount=8;
	int i;
	for(i=0;i<MAXSEGMENT;i++) {
		level.segment[i].isBoss=0;
		level.segment[i].isTunnel=0;
		level.segment[i].visualization=1;
		if(level.segment[i].heightMap) freeImage(level.segment[i].heightMap);
		level.segment[i].heightMap=0;
		if(level.segment[i].colorMap) freeImage(level.segment[i].colorMap);
		level.segment[i].colorMap=0;
		level.segment[i].cameraAngle=0;
		level.segment[i].seed[0]=1/6.0f;
		level.segment[i].seed[1]=1/5.0f;
		level.segment[i].scale=25;
		level.segment[i].length=32;
		level.segment[i].fogNear=128;
		level.segment[i].fogFar=1024;
		level.segment[i].fogColor=0xc0c0c0;
		level.segment[i].duration=level.duration/level.segmentCount;
		level.segment[i].elapsed=0;
		level.bossHealth=128;
		level.bossHealthMax=128;
		int p;
		for(p=0;p<256;p++) {
			int r,g,b;
			g=p<128?0:(p-128)+16;
			b=p<128?p+16:128+16;
			r=g;
			level.segment[i].palette[p]=r|(g<<8)|(b<<16);
		}
	}
	level.activeSegment=0;
	level.elapsed=0;
	level.segment[0].elapsed=0;
	triggerCount=0;
	bonusTriggerCount=0;
	enemyCount=0;
	enemyHit=0;
	enemyMiss=0;
	evoKills=0;
	chaosKills=0;
	supportHit=0;
	supportMiss=0;
	invulnerableTimer=5000;
	targeted=0;
	level.bpm=148;
	level.evoDropRate=8;
	level.chaosDropRate=20;
	level.initialEvo=0;
	level.initialChaos=1;
}

void loadLevel(const char *fname)
{
	FILE *file=fopen(fname,"r");
	
	if(!file) {
		setMode(MENUMODE);
		strcpy(alertMessage,"Couldn't find ");
		strcat(alertMessage,fname);
		return;
	}
	newLevel();
	strcpy(level.name,fname);
	int firstSegment=0;
	int activeSegment=0;
	int firstTrigger=0;
	int activeTrigger=0;
	char buf[256];
	unsigned int checkSumU=0;
	float checkSumF=0;
	int checkSumI=0;
	while(fgets(buf,255,file)) {
		buf[255]=0;
		if(buf[0]=='#') continue;
		if(strchr(buf,'\n')) strchr(buf,'\n')[0]=0;
		if(buf[0]=='[') {
			if(strcmp(buf,"[segment]")==0) {
				if(firstSegment++) activeSegment++;
				if(activeSegment>=MAXSEGMENT) activeSegment=0;
			} else if(strcmp(buf,"[trigger]")==0) {
				if(firstTrigger++) activeTrigger++;
				if(activeTrigger>=MAXTRIGGER) activeTrigger=0;
			}
		}
		char *value=strchr(buf,'=');
		if(value) {
			value[0]=0;
			value++;
			int i=0;
			float f=0;
			unsigned int u=0;
			sscanf(value,"%d",&i);
			sscanf(value,"%f",&f);
			sscanf(value,"0x%x",&u);
			checkSumF+=f;
			checkSumI+=i;
			checkSumU+=u;
			if(strcmp(buf,"segments")==0) {
				level.segmentCount=i;
			} else if(strcmp(buf,"duration")==0) {
				level.duration=f*60*1000;
			} else if(strcmp(buf,"bpm")==0) {
				if(f<1) f=1;
				level.bpm=f;
			} else if(strcmp(buf,"chaosDropRate")==0) {
				level.chaosDropRate=i;
			} else if(strcmp(buf,"evoDropRate")==0) {
				level.evoDropRate=i;
			} else if(strcmp(buf,"initialChaos")==0) {
				level.initialChaos=i;
			} else if(strcmp(buf,"initialEvo")==0) {
				level.initialEvo=i;
			} else if(strcmp(buf,"isBoss")==0) {
				level.segment[activeSegment].isBoss=i;
				printf("segment %d.isBoss==%d\n",activeSegment,i);
			} else if(strcmp(buf,"isTunnel")==0) {
				level.segment[activeSegment].isTunnel=i;
			} else if(strcmp(buf,"visualization")==0) {
				level.segment[activeSegment].visualization=i;
			} else if(strcmp(buf,"cameraAngle")==0) {
				level.segment[activeSegment].cameraAngle=i;
			} else if(strcmp(buf,"fogNear")==0) {
				level.segment[activeSegment].fogNear=f;
			} else if(strcmp(buf,"fogFar")==0) {
				level.segment[activeSegment].fogFar=f;
			} else if(strcmp(buf,"fogColor")==0) {
				level.segment[activeSegment].fogColor=u;
			} else if(strcmp(buf,"scale")==0) {
				level.segment[activeSegment].scale=f;
			} else if(strcmp(buf,"length")==0) {
				level.segment[activeSegment].length=f;
			} else if(strcmp(buf,"seedA")==0) {
				level.segment[activeSegment].seed[0]=f;
			} else if(strcmp(buf,"seedB")==0) {
				level.segment[activeSegment].seed[1]=f;
			} else if(strncmp(buf,"palette",7)==0) {
				int p=0;
				sscanf(buf,"palette%d",&p);
				if(p<0 || p>255) p=0;
				level.segment[activeSegment].palette[p]=u;
			} else if(strcmp(buf,"triggerType")==0) {
				strcpy(trigger[activeTrigger].type,value);
				findEnemyModel(value);
			} else if(strcmp(buf,"triggerSegment")==0) {
				trigger[activeTrigger].segment=i;
			} else if(strcmp(buf,"health")==0) {
				trigger[activeTrigger].health=i;
			} else if(strcmp(buf,"points")==0) {
				trigger[activeTrigger].points=i;
			} else if(strcmp(buf,"bonusType")==0) {
				trigger[activeTrigger].bonusType=i;
			} else if(strcmp(buf,"stageCount")==0) {
				trigger[activeTrigger].stageCount=i;
			} else if(strncmp(buf,"stage",5)==0) {
				int st=0;
				char thing[255];
				sscanf(buf,"stage%d%s",&st,thing);
				if(st<0 || st>7) st=0;
				if(strcmp(thing,"x")==0) {
					trigger[activeTrigger].stage[st].pos.x=f;
				} else if(strcmp(thing,"y")==0) {
					trigger[activeTrigger].stage[st].pos.y=f;
				} else if(strcmp(thing,"z")==0) {
					trigger[activeTrigger].stage[st].pos.z=f;
				} else if(strcmp(thing,"rx")==0) {
					trigger[activeTrigger].stage[st].orient.x=f*GU_PI/180;
				} else if(strcmp(thing,"ry")==0) {
					trigger[activeTrigger].stage[st].orient.y=f*GU_PI/180;
				} else if(strcmp(thing,"rz")==0) {
					trigger[activeTrigger].stage[st].orient.z=f*GU_PI/180;
				} else if(strcmp(thing,"trigger")==0) {
					trigger[activeTrigger].stage[st].trigger=f*level.duration/level.segmentCount;
				}			
			}
		}
	}
	triggerCount=activeTrigger;
	sprintf(level.checkSum,"%d:0x%06x:%f",checkSumI,checkSumU,checkSumF);
	int i;
	for(i=0;i<triggerCount;i++) trigger[i].actor=0;
	fclose(file);
	genTerrainNoPoly();
}

const char *getCheckSum()
{
	return level.checkSum;
}

const char *getName()
{
	return level.name;
}

void newGame()
{
	if(!reticleImage[0]) {
		loadOptions();
		reticleImage[0]=loadPng("data/base.png");
		swizzleFast(reticleImage[0]);
		int i;
		for(i=1;i<9;i++) {
			char fname[]="data/reticlen.png";
			fname[12]='0'+i;
			printf("Loading '%s'\n",fname);
			reticleImage[i]=loadPng(fname);
			swizzleFast(reticleImage[i]);
		}
	}
	if(!hudBossImage) {
		hudBossImage=loadPng("data/boss meter.png");
		swizzleFast(hudBossImage);
	}	
	if(!hudLeftImage) {
		hudLeftImage=loadPng("data/evolution bar.png");
		swizzleFast(hudLeftImage);
	}	
	if(!hudRightImage) {
		hudRightImage=loadPng("data/chaos bar.png");
		swizzleFast(hudRightImage);
	}
	if(!chaosImage) {
		chaosImage=loadPng("data/chaos.png");
		swizzleFast(chaosImage);
	}
	if(!heroModel[0].mesh[0]) loadAnimModel(heroModel+0,"models/hero/zooplankton.md2","models/hero/zooplankton.png");
	if(!heroModel[1].mesh[0]) loadAnimModel(heroModel+1,"models/hero/seaurchin.md2","models/hero/seaurchin.png");
	if(!heroModel[2].mesh[0]) loadAnimModel(heroModel+2,"models/hero/seahorse.md2","models/hero/seahorse.png");
	if(!heroModel[3].mesh[0]) loadAnimModel(heroModel+3,"models/hero/sting.md2","models/hero/sting.png");
	if(!heroModel[4].mesh[0]) loadAnimModel(heroModel+4,"models/hero/thresher.md2","models/hero/thresher.png");
	initActor(&hero,heroModel+1);
	hero.health=2;
	hero.requestedMotion=MT_IDLE;
	hero.frameDuration=1000/20;
	hero.pos.y=-35;
	hero.orient.y=M_PI;
	const char *enemyModelName[]={ "powerup/bonuslevelup.md2","powerup/bonuschaosup.md2","powerup/bonusevoup.md2","powerup/leveldown.md2","powerup/levelup.md2","powerup/chaosup.md2","powerup/evoup.md2"};
	int i;
	for(i=0;i<7;i++) {
		findEnemyModel(enemyModelName[i]);
	}
	loadItems();
	
	enemyCount=0;
	currentDir=0;
	from.x=155; from.y=155; from.z=155;
	to.x=0; to.y=0; to.z=0;
	up.x=0; up.y=1; up.z=0;
	fromTarget.x=155; fromTarget.y=155; fromTarget.z=155;
	updateCamera(0);
	reticle.x=240; reticle.y=131; reticle.z=0;
	alertMessage[0]=0;
//	visual=1;
//	visualRequestId++;
	triggerCount=0;
printf("Current mode is %d, new mode is to be %d\n",gameMode,newMode);
	if(gameMode==USERLEVELMODE) {
		loadLevel(userLevelList[userLevelFirst+menuItem]);
		printf("Loaded level '%s'\n",userLevelList[userLevelFirst+menuItem]);
	} else {
		loadLevel("map/level001.txt");
		printf("Loaded first level\n");
	}
	gameScore=0;
	shownScore=0;
	chaosUp=2;
	evoUp=6;
	chaosLeft=0;
	menuItem=0;
}

void drawTeather(struct Teather *teather)
{
	float p1[3],p2[3];
	float *p0,*p3;
	float w[3];
	p0=&teather->from.x;
	p3=&teather->to.x;
	p1[0]=p0[0]*(2/3.0f)+p3[0]*(1/3.0f);
	p1[1]=p0[1]+40;
	p1[2]=p0[2]*(2/3.0f)+p3[2]*(1/3.0f);
	p2[0]=p0[0]*(1/3.0f)+p3[0]*(2/3.0f);
	p2[1]=p3[1]+40;
	p2[2]=p0[2]*(1/3.0f)+p3[2]*(2/3.0f);
	w[0]=(p0[2]-p3[2]);
	w[1]=0;
	w[2]=-(p0[0]-p3[0]);
	float dist=sqrtf(w[0]*w[0]+w[2]*w[2]);
	if(dist<0.1f) dist=1.0f;
	w[0]=w[0]*2/dist;
	w[2]=w[2]*2/dist;
	//w[0]=0; w[1]=0; w[2]=0;
	
	if(!teather->vert) return;
	int i;
	teather->vert->x=p0[0];
	teather->vert->y=p0[1];
	teather->vert->z=p0[2];
	teather->vert->color=GU_RGBA(0xff,0xc0,0x00,0xc0);
	for(i=0;i<teather->vertCount/2;i++) {
		struct Vertex3DCP *v=teather->vert+(i*2)+1;
		float t=i/(float)(teather->vertCount/2-1);
		float t1=1.0f-t;
		// (1-t)^3*p0 + 3*(1-t)^2*t*p1 + 3*(1-t)*t^2*p2 + t^3*p3
		v->x=t1*t1*t1*p0[0]+3*t1*t1*t*p1[0]+3*t1*t*t*p2[0]+t*t*t*p3[0];
		v->y=t1*t1*t1*p0[1]+3*t1*t1*t*p1[1]+3*t1*t*t*p2[1]+t*t*t*p3[1];
		v->z=t1*t1*t1*p0[2]+3*t1*t1*t*p1[2]+3*t1*t*t*p2[2]+t*t*t*p3[2];
		v->color=GU_RGBA(0x00,0xc0,0xa0,0xc0);
		if(i<teather->vertCount-1) {
			v[1].x=v[0].x+w[0];
			v[1].y=v[0].y+w[1];
			v[1].z=v[0].z+w[2];
			v[1].color=GU_RGBA(0xff,0xc0,0x00,0xc0);
		}
		v->x-=w[0];
		v->y-=w[1];
		v->z-=w[2];
	}
	sceKernelDcacheWritebackAll();
	long age=level.segment[level.activeSegment].elapsed-teather->when;
	sceGuDisable(GU_TEXTURE_2D);
	enum CameraAngle cameraAngle=level.segment[level.activeSegment].cameraAngle;
	if(age<500) {
		int k;
		for(k=1;k<4;k++) {
			float w=10+128*(500-age)/500/k;
			struct Vertex3DCP *vert=(struct Vertex3DCP *)sceGuGetMemory(sizeof(struct Vertex3DCP)*8),*v;
			v=vert;
			int a=(255-128*age/500)/k;
			float per=age/500.0f,cent=1-per;
			
			for(i=0;i<2;i++) {
				int j;
				for(j=0;j<4;j++) {
					float theta=(GU_PI*0.5)*(j+i)+(GU_PI*0.5f/500)*age+(GU_PI/8)*k;
					v->x=teather->from.x*per+teather->to.x*cent+(cameraAngle==CA_FORWARD?(cosf(theta)*w):0);
					v->y=teather->from.x*per+teather->to.y*cent+sinf(theta)*w;
					v->z=teather->from.z*per+teather->to.z*cent+(cameraAngle!=CA_FORWARD?(cosf(theta)*w):0);
					v->color=GU_RGBA(255,255,255,a);
					v++;
				}
			}
			sceGumDrawArray(GU_LINES,GU_VERTEX_32BITF|GU_COLOR_8888|GU_TRANSFORM_3D,8,0,vert);
		}
	}
	sceGuDisable(GU_CULL_FACE);
	sceGumDrawArray(GU_TRIANGLE_STRIP,GU_VERTEX_32BITF|GU_COLOR_8888|GU_TRANSFORM_3D,teather->vertCount,0,teather->vert);
	sceGuEnable(GU_CULL_FACE);
	sceGuEnable(GU_TEXTURE_2D);
}

#if 0
// Depth sort function
float A,B,C,D;
void updateDepthSort()
{
	A=to.x-from.x;
	B=to.y-from.y;
	C=to.z-from.z;
	float dist=sqrtf(A*A+B*B+C*C);
	if(dist>0) dist=1.0f/dist;
	A*=dist;
	B*=dist;
	C*=dist;
	D=-(from.x*A+from.y*B+from.z*C);
}

int depthCmp(const void *one,const void *two)
{
	struct Vertex3DTP *a=(struct Vertex3DTP *)one;
	struct Vertex3DTP *b=(struct Vertex3DTP *)two;
	
	float x=(a[0].x+a[1].x+a[2].x); //*(1/3.0f);
	float y=(a[0].y+a[1].y+a[2].y); //*(1/3.0f);
	float z=(a[0].z+a[1].z+a[2].z); //*(1/3.0f);
	float d1=x*A+y*B+z*C; //+D;
	x=(b[0].x+b[1].x+b[2].x); //*(1/3.0f);
	y=(b[0].y+b[1].y+b[2].y); //*(1/3.0f);
	z=(b[0].z+b[1].z+b[2].z); //*(1/3.0f);
	float d2=x*A+y*B+z*C; //+D;
	return d1<d2?-1:d1>d2?1:0;
}
#endif

void drawActor(struct Actor *actor)
{
	// Save the context, do the transform, draw the meshes restore the context	
	if(!actor->active) return;
	struct AnimModel *model=actor->model;
	struct AnimMesh *mesh=model->mesh[0];
	if(!mesh) {
		printf("Actor @ (%.2f,%.2f,%.2f) model missing a mesh.\n",actor->pos.x,actor->pos.y,actor->pos.z);
		return;
	}
	float fraction=(actor->frameElapsed%actor->frameDuration)/(float)actor->frameDuration;
	//printf("Fraction: %.3f;",fraction);
	int frame,nextFrame;
	frame=actor->lastFrame;
	nextFrame=actor->frame;
	
	struct Vertex3DTP *buffer=sceGuGetMemory( sizeof(struct Vertex3DTP)*mesh->polyCount*3);
	//if(model->cullBackface==0) buffer=(struct Vertex3DTP *)(0x0fffffff&(int)buffer);
	md2GetFractionalFrame(mesh,frame,nextFrame,fraction,buffer,actor->minMax);
	Image *source=mesh->texture;
	if(source) {
		sceGuTexMode(GU_PSM_8888, 0, 0, source->isSwizzled);
		sceGuTexImage(0, source->textureWidth, source->textureHeight, source->textureWidth, source->data);
	}
	sceGumPushMatrix();
	sceGumTranslate(&actor->pos);
	if(actor->orient.x!=0 || actor->orient.y!=0 || actor->orient.z!=0) {
		sceGumRotateXYZ(&actor->orient);
	}
	if(model->cullBackface==0) {
		sceGuDisable(GU_CULL_FACE);
		//updateDepthSort();
		//qsort(buffer,actor->model->mesh[0]->polyCount,sizeof(struct Vertex3DTP)*3,depthCmp);
		sceKernelDcacheWritebackAll();
	}
	sceGumDrawArray(GU_TRIANGLES,GU_VERTEX_32BITF|GU_TEXTURE_32BITF|GU_TRANSFORM_3D,actor->model->mesh[0]->polyCount*3,0,buffer);
	if(model->cullBackface==0) sceGuEnable(GU_CULL_FACE);

#if 0
	if(showFps) {
		struct Vertex3DCP *vert=sceGuGetMemory(sizeof(struct Vertex3DCP)*12*2),*v;
		v=vert;
		const int corners[12][6]={ {0,1,2,3,1,2},{0,1,2,0,4,2},{0,1,2,0,1,5},{3,4,5,0,4,5},{3,4,5,3,1,5},{3,4,5,3,4,2},{3,1,2,3,4,2},{3,4,2,0,4,2},{0,4,2,0,4,5},{3,1,5,0,1,5},{3,1,5,3,1,2},{0,1,5,0,4,5}, };
		int i;
		for(i=0;i<12;i++) {
			v->x=actor->minMax[corners[i][0]];
			v->y=actor->minMax[corners[i][1]];
			v->z=actor->minMax[corners[i][2]];
			v->color=GU_RGBA(255,255,255,255);
			v++;
			v->x=actor->minMax[corners[i][3]];
			v->y=actor->minMax[corners[i][4]];
			v->z=actor->minMax[corners[i][5]];
			v->color=GU_RGBA(255,255,255,255);
			v++;
		}
		sceGuDisable(GU_TEXTURE_2D);
		sceGumDrawArray(GU_LINES,GU_VERTEX_32BITF|GU_COLOR_8888|GU_TRANSFORM_3D,12*2,0,vert);
		sceGuEnable(GU_TEXTURE_2D);
	}
#endif

	//printf("drawActor: f=%d,last f=%d, portion=%d\n",hero.frame,hero.lastFrame,hero.frameElapsed);
	static Image *locked=0;
	if(!locked) {
		locked=loadPng("data/locked.png");
		swizzleFast(locked);
	}
	if(actor->targeted && locked) {
		struct Vertex3DTP *v=sceGuGetMemory(sizeof(struct Vertex3DTP)*4);
		float x,y;
		x=locked->imageWidth;
		y=-locked->imageHeight;
		switch(level.segment[level.activeSegment].cameraAngle) {
		case CA_LEFT:
			sceGumRotateY((float)M_PI/2);
			break;
		case CA_RIGHT:
			sceGumRotateY(-(float)M_PI/2);
			break;
		default:
			break;
		}
		x*=0.5f;
		y*=0.5f;
		v[0].x=x;
		v[0].y=y;
		v[0].z=0;
		v[0].u=0;
		v[0].v=0;
		v[1].x=x;
		v[1].y=-y;
		v[1].z=0;
		v[1].u=1;
		v[1].v=0;
		v[2].x=-x;
		v[2].y=-y;
		v[2].z=0;
		v[2].u=1;
		v[2].v=1;
		v[3].x=-x;
		v[3].y=y;
		v[3].z=0;
		v[3].u=0;
		v[3].v=1;
		source=locked;
		if(source) {
			sceGuTexMode(GU_PSM_8888, 0, 0, source->isSwizzled);
			sceGuTexImage(0, source->textureWidth, source->textureHeight, source->textureWidth, source->data);
			int i;
			for(i=0;i<4;i++) {
				if(v[i].u==1) v[i].u=source->imageWidth/(float)source->textureWidth;
				if(v[i].v==1) v[i].v=source->imageHeight/(float)source->textureHeight;
			}
		}
		sceGumDrawArray(GU_TRIANGLE_FAN,GU_VERTEX_32BITF|GU_TEXTURE_32BITF|GU_TRANSFORM_3D,4,0,v);
	}
	sceGumPopMatrix();
	if(actor->targeted && actor->teather.vert) drawTeather(&actor->teather);
}

void drawExplosion(struct Explosion *explosion)
{
	if(explosion->elapsed<1) return;
	sceGumMatrixMode(GU_MODEL);
	sceGumPushMatrix();
	sceGumTranslate(&explosion->pos);
	const int particleCount=128;
	Image *source=0;
	
	struct Vertex3DCP *vert=(struct Vertex3DCP *)explosion->vert,*v=0;
	if(!vert) {
		vert=(struct Vertex3DCP *)malloc(sizeof(struct Vertex3DCP)*particleCount*2);
		explosion->vert=(struct Vertex3DCNP *)vert;
		explosion->stripCount=1;
		explosion->vertCount=particleCount*2;
	}
	v=vert;
	unsigned int r,g,b;
	unsigned int a=255-(int)(255.0f*explosion->elapsed/(float)explosion->duration);

	float radius=explosion->elapsed/(float)explosion->duration*60.0f+10.0f;
	int i,j;
	
	for(i=0;i<particleCount;i++) {
		// Spreading spiral
		float theta=i*GU_PI*8.0f/particleCount;
		float phi=i*GU_PI/particleCount-(GU_PI/2);
		r=explosion->firstColor&0xff;
		g=(explosion->firstColor>>8)&0xff;
		b=(explosion->firstColor>>16)&0xff;
		int offset=((int)explosion->elapsed)%explosion->duration;
		float per=1.0f-(offset/(float)explosion->duration);
		per=per+(i%9)/10.0f;
		if(per>1.0f) per-=1.0f;
		float cent=1.0f-per;
		r=(unsigned int)(r*per+(explosion->lastColor&0xff)*cent);
		g=(unsigned int)(g*per+((explosion->lastColor>>8)&0xff)*cent);
		b=(unsigned int)(b*per+((explosion->lastColor>>16)&0xff)*cent);
		unsigned int color=r|(g<<8)|(b<<16)|(a<<24);
		float x=sinf(phi)*cosf(theta)*radius;
		float y=sinf(phi)*sinf(theta)*radius;
		float z=cosf(phi)*radius;
		for(j=-1;j<=1;j+=2) {
			v->x=x+(view.x.x+view.x.y)*j*16;
			v->y=y+(view.y.x+view.y.y)*j*16;
			v->z=z+(view.z.x+view.z.y)*j*16;
			v->color=color;
			v++;
		}
	}
	// Now sort it by distance from viewer.
	//qsort(vert,particleCount,sizeof(struct Vertex3DCP)*2,spriteViewDistCmp);
	sceKernelDcacheWritebackAll();
	sceGuBlendFunc(GU_ADD,GU_FIX,GU_FIX,0xffffffff,0xffffffff);
	if(source) {
		sceGuTexMode(GU_PSM_8888, 0, 0, source->isSwizzled);
		sceGuTexImage(0, source->textureWidth, source->textureHeight, source->textureWidth, source->data);
	}
	sceGuDisable(GU_DEPTH_TEST);
	sceGumDrawArray(GU_SPRITES,GU_VERTEX_32BITF|GU_COLOR_8888|GU_TRANSFORM_3D,explosion->vertCount,0,vert);
	sceGuBlendFunc(GU_ADD,GU_SRC_ALPHA,GU_ONE_MINUS_SRC_ALPHA,0,0);
	sceGuEnable(GU_DEPTH_TEST);
	sceGumPopMatrix();
}

void drawChaos()
{
	if(chaosLeft<1) return;
	sceGumMatrixMode(GU_MODEL);
	sceGumPushMatrix();
	sceGumTranslate(&hero.pos);
	const int particleCount=128;
	const int vertCount=2*particleCount;
	Image *source=chaosImage;
	
//	struct Vertex3DTCP *vert=(struct Vertex3DTCP *)sceGuGetMemory(sizeof(struct Vertex3DTCP)*vertCount),*v;
	static struct Vertex3DTCP *vert=0;
	if(!vert) vert=(struct Vertex3DTCP *)malloc(sizeof(struct Vertex3DTCP)*vertCount);
	struct Vertex3DTCP *v;
	v=vert;
	unsigned int r,g,b;
	int elapsed=chaosMax-chaosLeft;
	unsigned int a=12-(int)(12.0f*elapsed/(float)chaosMax);

	float radius=elapsed/(float)chaosMax*256.0f+10.0f;
	int i,j;

	unsigned int firstColor=GU_RGBA(128,255,128,64),lastColor=GU_RGBA(0,255,0,64);
	
	for(i=0;i<particleCount;i++) {
		// Spreading spiral column
		//float theta=i*(float)M_PI*8.0f/particleCount;
		r=firstColor&0xff;
		g=(firstColor>>8)&0xff;
		b=(firstColor>>16)&0xff;
		int offset=((int)elapsed)%chaosMax;
		float per=1.0f-(offset/(float)chaosMax);
		per=per+(i%9)/10.0f;
		if(per>1.0f) per-=1.0f;
		float cent=1.0f-per;
		r=(unsigned int)(r*per+(lastColor&0xff)*cent);
		g=(unsigned int)(g*per+((lastColor>>8)&0xff)*cent);
		b=(unsigned int)(b*per+((lastColor>>16)&0xff)*cent);
		unsigned int color=r|(g<<8)|(b<<16)|(a<<24);
		float phi=(i/2)*(GU_PI*2/(particleCount/2))*(i%2?-1:1);
		float x=sinf(phi)*radius;
		float y=(i-particleCount/2)*radius/particleCount;
		if(y<0) y=-y;
		y=(y-radius/4)*3;
		float z=cosf(phi)*radius;
		for(j=-1;j<=1;j+=2) {
			v->x=x+(view.x.x+view.x.y)*j*32;
			v->y=y+(view.y.x+view.y.y)*j*32;
			v->z=z+(view.z.x+view.z.y)*j*32;
			v->color=color;
			v->u=(j==1?0.25f:0)+0.25f*((i/2)%4);
			v->v=(j==1?0.25f:0)+0.5f;
			v++;
		}
	}
	if(source) {
		sceGuEnable(GU_TEXTURE_2D);
		sceGuTexMode(GU_PSM_8888, 0, 0, source->isSwizzled);
		sceGuTexImage(0, source->textureWidth, source->textureHeight, source->textureWidth, source->data);
	}
	//sceGuTexWrap(GU_CLAMP,GU_CLAMP);
	sceGuDisable(GU_DEPTH_TEST);
	sceKernelDcacheWritebackAll();
	sceGumDrawArray(GU_SPRITES,GU_VERTEX_32BITF|GU_TEXTURE_32BITF|GU_COLOR_8888|GU_TRANSFORM_3D,vertCount,0,vert);
	//sceGuTexWrap(GU_REPEAT,GU_REPEAT);
	sceGuEnable(GU_DEPTH_TEST);
	sceGumPopMatrix();
	if(source) {
		sceGuDisable(GU_TEXTURE_2D);
	}
}

int terrainDebug=1;
unsigned short *terrainPoly;
struct Vertex3DCNP *terrainVert;
const int grid=128;
//float scale=32;
Image *currentImage,*colorMap;
unsigned int terrainPalette[256];

void genTerrainNoPoly()
{
	int i,j,k;
	int tile=256/grid;
	if(terrainVert) free(terrainVert);
	terrainVert=0;
	struct Vertex3DCNP *vert=malloc(sizeof(struct Vertex3DCNP)*grid*grid*2);
	terrainVert=vert;
	if(!currentImage) currentImage=newImage(256,256);
//	if(!nextImage) nextImage=newImage(grid,grid);
	if(terrainPalette[255]==0) {
		for(i=0;i<255;i++) terrainPalette[i]=i|(i<<8)|(i<<16)|(255<<24);
		copyPalette(terrainPalette);
	}
	if(colorMap) freeImage(colorMap);
	colorMap=0;
	if(level.segment[level.activeSegment].visualization!=4) {
		generateVisualization(level.segment[level.activeSegment].seed[0],level.segment[level.activeSegment].seed[1],currentImage,level.segment[level.activeSegment].visualization);
	} else {
		char buf[256];
		strcpy(buf,level.name);
		char *s=buf+strlen(buf);
		if(strrchr(buf,'.')) s=strrchr(buf,'.');
		sprintf(s,"seg%dheight.png",level.activeSegment);
		Image *image=loadPng(buf);
		if(image && image->imageWidth==256 && image->imageHeight==256) {
			freeImage(currentImage);
			currentImage=image;
		} else {
			printf("Could not find 256x256 hight map called '%s'\n",buf);
		}
		sprintf(s,"seg%dcolor.png",level.activeSegment);
		colorMap=loadPng(buf);
		if(colorMap && (colorMap->imageWidth!=256 || colorMap->imageHeight!=256)) {
			printf("colorMap file '%s' not 256x256! Discarded.\n",buf);
			freeImage(colorMap);
			colorMap=0;
		}
	}
//	if(level.activeSegment+1<level.segmentCount) generateVisualization(level.segment[level.activeSegment+1].seed[0],level.segment[level.activeSegment+1].seed[1],nextImage,level.segment[level.activeSegment+1].visualization);
	for(i=0;i<grid-1;i++) {
		for(j=0;j<grid;j++) {
			for(k=0;k<2;k++) {
				struct Vertex3DCNP *v=vert+(i*grid+j)*2+k;
				unsigned int color=currentImage->data[tile*(i+k)*currentImage->textureWidth+tile*j];
				unsigned int terrainZ=(color&0xff);
				if(!colorMap) {
					color=level.segment[level.activeSegment].palette[terrainZ];
				} else {
					color=colorMap->data[tile*(i+k)*currentImage->textureWidth+tile*j];
				}
				//if(terrainDebug) printf("%c",'a'+(terrainZ*26/255));
				v->color=color | GU_RGBA(0,0,0,255); //GU_RGBA(0,terrainZ,terrainZ,255);
				v->x=((i+k)-grid/2)*tile*level.segment[level.activeSegment].length;
				v->z=(j)*tile*level.segment[level.activeSegment].length;
				v->y=((int)terrainZ-255)*(1/255.0f)*level.segment[level.activeSegment].scale-155;
				if(level.segment[level.activeSegment].isTunnel!=0) {
					// Transform x and y into a circle.
					float angle=(i+k+grid/4)/(float)(grid-2)*3.1415926535f*2;
					float y=v->y;
					v->x=cosf(angle)*(25-y);
					v->y=sinf(angle)*(25-y);
				}
				if(i>0 && j>0 && i<grid-1 && j<grid-1) {
					ScePspFVector3 v1,v2;
					v1.x=v->x-vert[(i*grid+j-1)*2+k].x;
					v1.y=v->y-vert[(i*grid+j-1)*2+k].y;
					v1.z=v->z-vert[(i*grid+j-1)*2+k].z;
					v2.x=v->x-vert[((i-1)*grid+j)*2+k].x;
					v2.y=v->y-vert[((i-1)*grid+j)*2+k].y;
					v2.z=v->z-vert[((i-1)*grid+j)*2+k].z;
					gumCrossProduct((ScePspFVector3 *)&v->nx,&v1,&v2);
					gumNormalize((ScePspFVector3 *)&v->nx);
				} else {
					v->nx=0;
					v->ny=1;
					v->nz=0;
				}
			}
		}
		//if(terrainDebug) printf("\n");
	}
	sceKernelDcacheWritebackAll();
}

void drawTerrain()
{
	// Now take a piece of it and draw it.
	struct Vertex3DCNP *vert=terrainVert;
	unsigned short *poly=terrainPoly;
	sceGumMatrixMode(GU_MODEL);
	sceGumPushMatrix();
	//sceGumRotateXYZ(&rot);
	ScePspFVector3 scale={1,0,1};
	scale.y=1+0.03f*cosf(level.elapsed*(GU_PI*2/60000*level.bpm));
	sceGumScale(&scale);
	if(!vert) {
		genTerrainNoPoly();
		vert=terrainVert;
		poly=terrainPoly;
	}
	terrainDebug=0;
	//sceGumDrawSpline(GU_VERTEX_32BITF|GU_COLOR_8888|GU_NORMAL_32BITF|GU_TRANSFORM_3D,32,32,3,3,0,vert);
	float levelLength=level.segment[level.activeSegment].length;
	float levelDuration=level.segment[level.activeSegment].duration;
	float levelElapsed=level.segment[level.activeSegment].elapsed;
	int delta=levelDuration-levelElapsed;
	int i;
	int many=grid*2;
	if(many*levelLength-250>1024) {
		int newMany=(int)(delta/(float)levelDuration*grid*2+250/levelLength);
		if(newMany<many) {
			if(newMany>3) many=newMany; else many=3;
		}
	}
	int start=0;
	if(many<grid*2) {
		if(many&1) many++;
		start=grid*2-many;
	}
	for(i=2;i<grid-1;i++) {
		sceGumDrawArray(GU_TRIANGLE_STRIP,GU_VERTEX_32BITF|GU_COLOR_8888|GU_NORMAL_32BITF|GU_TRANSFORM_3D,many,0,vert+grid*2*i+start);
	}
	if((float)delta/levelDuration*levelLength*256<1024) {
		sceGumPushMatrix();
		ScePspFVector3 trans={0,0,256*levelLength};
		sceGumTranslate(&trans);
		for(i=2;i<grid-1;i++) {
			struct Vertex3DCNP *bridge=(struct Vertex3DCNP *)sceGuGetMemory(sizeof(struct Vertex3DCNP)*4);
			bridge[0]=vert[grid*2*(i+1)-2];
			bridge[1]=vert[grid*2*(i+1)-1];
			bridge[2]=vert[grid*2*i];
			bridge[3]=vert[grid*2*i+1];
			bridge[0].z-=trans.z;
			bridge[1].z-=trans.z;
			sceGumDrawArray(GU_TRIANGLE_STRIP,GU_VERTEX_32BITF|GU_COLOR_8888|GU_NORMAL_32BITF|GU_TRANSFORM_3D,4,0,bridge);
			sceGumDrawArray(GU_TRIANGLE_STRIP,GU_VERTEX_32BITF|GU_COLOR_8888|GU_NORMAL_32BITF|GU_TRANSFORM_3D,start,0,vert+grid*2*i);
		}
		sceGumPopMatrix();
	}
	if(delta<5000) {	// end of segment is coming...
#if 0
		ScePspFVector3 pos={0,32,256*level.segment[level.activeSegment].length*(0.0f+delta/5000.0f)};
		sceGumTranslate(&pos);
		//printf("%d: hero.pos.z=%f; pos.z=%f\n",delta,hero.pos.z,pos.z);
		for(i=2;i<grid-1;i++) {
			if(poly) {
				sceGumDrawArray(GU_LINES,GU_VERTEX_32BITF|GU_COLOR_8888|GU_NORMAL_32BITF|GU_INDEX_16BIT|GU_TRANSFORM_3D,(grid-1)*2+1,poly+i*((grid-1)*2+1),vert);
			} else {
				sceGumDrawArray(GU_LINES,GU_VERTEX_32BITF|GU_COLOR_8888|GU_NORMAL_32BITF|GU_TRANSFORM_3D,grid*2,0,vert+grid*2*i);
			}
		}
#else
		sceGuDisable(GU_DEPTH_TEST);
		sceGuDisable(GU_LIGHTING);
		sceGuDisable(GU_CULL_FACE);
		static struct Vertex3DCPfast {
			u32 color;
			float x,y,z;
		} *tr=0;
		const int trCount=128;
		if(!tr) tr=malloc(sizeof(struct Vertex3DCPfast)*trCount*2);
		int alpha=255-255*delta/5000;
		for(i=0;i<trCount*2;i+=2) {
			float angle=i*(GU_PI/(trCount-1));
			float c=cosf(angle),s=sinf(angle);
			u32 rgba=i%4?GU_RGBA(255,255,255,alpha):GU_RGBA(192,192,192,alpha);
			tr[i].color=rgba;
			tr[i].x=c*delta/20;
			tr[i].y=s*delta/20;
			tr[i].z=hero.pos.z;
			angle+=(delta-2500)/2500.0f;
			c=cosf(angle);
			s=sinf(angle);
			tr[i+1].color=rgba;
			tr[i+1].x=c*260;
			tr[i+1].y=s*260;
			tr[i+1].z=hero.pos.z;
		}
		sceKernelDcacheWritebackAll();
		sceGumDrawArray(GU_TRIANGLE_STRIP,GU_VERTEX_32BITF|GU_COLOR_8888|GU_TRANSFORM_3D,trCount*2,0,tr);
		sceGuEnable(GU_CULL_FACE);
		sceGuEnable(GU_LIGHTING);
		sceGuEnable(GU_DEPTH_TEST);
#endif
	}
	
	sceGumPopMatrix();	
}

void draw()
{
	u64 oldTick,newTick;
	int i;
	
	pspDebugScreenSetXY(0,1);
	sceGuStart(GU_DIRECT,gulist);
	// clear screen
	sceRtcGetCurrentTick(&oldTick);
	sceGuClearColor(level.segment[level.activeSegment].fogColor|GU_RGBA(0,0,0,255));
	sceGuClearDepth(0);
	sceGuClear(GU_COLOR_BUFFER_BIT|GU_DEPTH_BUFFER_BIT);

	gumLoadIdentity(&view);
	gumLookAt(&view,&from,&to,&up);
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	sceGumLookAt((ScePspFVector3 *)&from,(ScePspFVector3 *)&to,(ScePspFVector3 *)&up);
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	sceGuFrontFace(GU_CW);

	sceGuFog(level.segment[level.activeSegment].fogNear,level.segment[level.activeSegment].fogFar,level.segment[level.activeSegment].fogColor);
	sceGuEnable(GU_FOG);
	sceGuDisable(GU_TEXTURE_2D);
	sceGuEnable(GU_DEPTH_TEST);
	sceGuEnable(GU_LIGHTING);

	// Draw terrain
	sceGumMatrixMode(GU_MODEL);
	sceGumPushMatrix();
	if(level.segment[level.activeSegment].isBoss) {
//		sceGumRotateY(level.elapsed/(6000.0f*GU_PI*2));
	}
	drawTerrain();
	sceGumMatrixMode(GU_MODEL);
	sceGumPopMatrix();
	// Draw decorations
	sceGuDisable(GU_LIGHTING);
	sceGuFrontFace(GU_CW);
	sceGuEnable(GU_TEXTURE_2D);
	sceGuEnable(GU_BLEND);
	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
	// Draw enemies
	for(i=0;i<enemyCount;i++) {
		drawActor(enemy+i);
	}
	// Draw boss
	if(level.segment[level.activeSegment].isBoss) {
		int i=-1;
		if(level.bossHealth>level.bossHealthMax/4*3) i=0;
		else if(level.bossHealth>level.bossHealthMax/4*2) i=1;
		else if(level.bossHealth>level.bossHealthMax/4*1) i=2;
		else if(level.bossHealth>0) i=3;
		if(i>=0) {
			ScePspFVector3 pos={hero.pos.x,hero.pos.y-120,hero.pos.z+256};
			if(level.segment[level.activeSegment].bossHealth<0) {
				pos.x-=(level.elapsed-level.bossTimer)/40;
			}
			setItemPos(i,pos.x,pos.y,pos.z);
			renderItem(i);
		}
		sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
	}
	// Draw hero
	drawActor(&hero);
	sceGuDisable(GU_FOG);

	sceGuEnable(GU_LIGHTING);
	sceGuDisable(GU_TEXTURE_2D);
	// Draw explosions
	for(i=0;i<explosionCount;i++) {
		if(explosion[i].active) drawExplosion(explosion+i);
	}
	drawChaos();
	
	sceGuEnable(GU_TEXTURE_2D);
	// Draw menu
	if(gameMode!=GAMEMODE) {
		drawMenu();
	} else {	// Draw hud
		int i;
		int segmentElapsed=level.segment[level.activeSegment].elapsed;
		if(segmentElapsed<500) {
			for(i=0;i<30;i++) {
				int w=20+i*(500-segmentElapsed)*(480/30)/500;
				drawFilledRect(reticle.x-w,reticle.y-w,w*2,1,GU_RGBA(255,255,255,255));
				drawFilledRect(reticle.x-w,reticle.y-w,1,w*2,GU_RGBA(255,255,255,255));
				drawFilledRect(reticle.x-w,reticle.y+w,w*2,1,GU_RGBA(255,255,255,255));
				drawFilledRect(reticle.x+w,reticle.y-w,1,w*2,GU_RGBA(255,255,255,255));
			}
		}
		
		for(i=0;i<8;i++) {
			drawFilledRect(9+(i%2)*8,184+i*8,9,9,evoUp>i?GU_RGBA(192,255,25,255):GU_RGBA(25,25,25,255));
		}
		for(i=0;i<4;i++) {
			drawFilledRect(454,184+i*17,7,14,chaosUp>i?GU_RGBA(128,25,25,255):GU_RGBA(25,25,25,255));
		}
		drawFilledRect(467,184,8,65,GU_RGBA(25,25,25,255));
		drawFilledRect(467,184,8,65*chaosLeft/chaosMax,GU_RGBA(25,128,25,255));
		if(level.segment[level.activeSegment].isBoss) {
			int lx=0,lw=0;
			if(hudBossImage) {
				lx=240-hudBossImage->imageWidth/2;
				lw=hudBossImage->imageWidth;
			}
			if(lw<=8) { lw=128; lx=240-64; }
			drawFilledRect(lx+4,24,lw-8,6,GU_RGBA(25,25,25,255));
			drawFilledRect(lx+4,24,(lw-8)*level.bossHealth/level.bossHealthMax,6,GU_RGBA(128,25,25,255));
			if(hudBossImage) drawSprite(0,0,hudBossImage->imageWidth,hudBossImage->imageHeight,hudBossImage,lx,20);
		}
		if(hudLeftImage) drawSprite(0,0,hudLeftImage->imageWidth,hudLeftImage->imageHeight,hudLeftImage,5,180);
		if(hudRightImage) drawSprite(0,0,hudRightImage->imageWidth,hudRightImage->imageHeight,hudRightImage,450,180);
		if(reticleImage[targeted]) drawSprite(0,0,40,40,reticleImage[targeted],reticle.x-20,reticle.y-20);
		if(shownScore<gameScore-200) shownScore+=21;
		else if(shownScore<gameScore-10) shownScore+=3;
		else if(shownScore<gameScore) shownScore+=1;
		if(showScore) {
			char buf[64];
			sprintf(buf,"%08d",shownScore);
			int w=0,h=0;
			extentMessage(&w,&h,FONT_SMALL,buf);
			drawMessage(480-4-w,4,FONT_SMALL,buf);
			int y=4+h;
			if(shownScore<gameScore && scoreBonus>0) {
				sprintf(buf,"Bonus %d%%",scoreBonus);
				extentMessage(&w,&h,FONT_SMALL,buf);
				drawMessage(480-4-w,y,FONT_SMALL,buf);
			}
			if(level.segment[level.activeSegment].isBoss) {
				sprintf(buf,"Boss: %d",level.bossHealth);
				y+=h;
				
				extentMessage(&w,&h,FONT_SMALL,buf);
				drawMessage(480-4-w,y,FONT_SMALL,buf);
			}
		}
	}
	
	sceGuFinish();
	sceGuSync(0,0);
	if(showFps) {
		sceRtcGetCurrentTick(&newTick);
		pspDebugScreenPrintf("FPS: %.2f (%.2f)",1000000.0f/(newTick-oldTick),((newTick-oldTick)/1000.0f));
		pspDebugScreenSetXY(20,1);
		if(level.segment[level.activeSegment].duration>0)
			pspDebugScreenPrintf("Seg %d; %.1f%%",level.activeSegment,100.0f*level.segment[level.activeSegment].elapsed/level.segment[level.activeSegment].duration);
	}
		
	sceDisplayWaitVblankStart();
	drawBuffer=sceGuSwapBuffers();
	pspDebugScreenSetOffset((int)drawBuffer);
	//printf("Draw buffer = %d\n",(int)drawBuffer);
}

void updateActor(struct Actor *actor,unsigned long elapsed,int mapPos)
{
	if(!actor->active) return;

	if(actor->trigger) {
		// Now work on interpolating the position and rotation.	
		if(actor->trigger->segment!=level.activeSegment) {	// Really couldn't happen.
			actor->trigger->actor=0;
			actor->active=0;
			return;
		}
		
		int j;
		// Now draw it where it would really be.
		struct Trigger *trigger=actor->trigger;
		for(j=0;j<trigger->stageCount-1;j++) {
			if(mapPos<trigger->stage[j].trigger) continue;
			if(mapPos>trigger->stage[j+1].trigger) continue;
			// in the range.
			break;
		}
		if(j<trigger->stageCount-1) {
			float per=1,cent=0;
			float delta=trigger->stage[j+1].trigger-trigger->stage[j].trigger;
			if(delta>0) {
				cent=(mapPos-trigger->stage[j].trigger)/delta;
				per=1.0f-cent;
			}
			actor->pos.x=trigger->stage[j].pos.x*per+trigger->stage[j+1].pos.x*cent;
			actor->pos.y=trigger->stage[j].pos.y*per+trigger->stage[j+1].pos.y*cent;
			actor->pos.z=trigger->stage[j].pos.z*per+trigger->stage[j+1].pos.z*cent;
			actor->orient.x=trigger->stage[j].orient.x*per+trigger->stage[j+1].orient.x*cent;
			actor->orient.y=trigger->stage[j].orient.y*per+trigger->stage[j+1].orient.y*cent;
			actor->orient.z=trigger->stage[j].orient.z*per+trigger->stage[j+1].orient.z*cent;
		} else {
			// mark as inactive, for deletion.
			if(actor->targeted) return triggerExplosion(actor);
			actor->active=0;
			actor->trigger->actor=0;
			actor->trigger=0;
			if(gameMode==GAMEMODE && actor->enemyType==ET_ACTIVE) {
				enemyMiss++;
				//evoKills=0;
			} else if(gameMode==GAMEMODE && (actor->enemyType==ET_CHAOSUP || actor->enemyType==ET_EVOUP)) {
				supportMiss++;
			}
			return;
		}
	}

	// animate the MD2
	if(actor->activeMotion!=MT_INACTIVE) {
		actor->frameElapsed+=elapsed;
		int frameMod=(actor->frameElapsed/actor->frameDuration);
		struct AnimMotion *motion=&actor->model->motion[actor->activeMotion];
		if(motion->loopLength>0) frameMod%=motion->loopLength; else frameMod=0;
		frameMod+=motion->startFrame;
		if(frameMod>motion->loopFrame) {
			frameMod-=motion->loopFrame;
			if(motion->loopLength<2) frameMod=0;
			else frameMod=frameMod%motion->loopLength;
			frameMod+=motion->loopFrame;
		}
		if(motion->loopLength>=2) {
			if(frameMod==actor->frame) return;
		}
		actor->lastFrame=actor->frame;
		actor->frame=frameMod;
		if(actor->requestedMotion==MT_INACTIVE) return;	// Transition time?
	} else {
		// first motion of the series.
		if(actor->requestedMotion==MT_INACTIVE) return;
		actor->lastFrame=actor->model->motion[actor->activeMotion].startFrame;
	}
	actor->activeMotion=actor->requestedMotion;
	actor->requestedMotion=MT_INACTIVE;
	actor->frame=actor->model->motion[actor->activeMotion].startFrame;
	actor->frameElapsed=0;
}

void updateTrigger(int mapPos)
{
	int i;
	// Enemies are squares
	if(level.segment[level.activeSegment].isBoss && level.segment[level.activeSegment].bossHealth<0) return;
	for(i=1;i<=triggerCount;i++) {
//		printf("trigger %d: seg %d, trigger=%d; mapPos=%d\n",i,trigger[i].segment,(int)trigger[i].stage[0].trigger,mapPos);
		if(trigger[i].segment!=level.activeSegment) continue;
		if(trigger[i].actor!=0) continue;
	
		int j;
		// Now draw it where it would really be.
		for(j=0;j<trigger[i].stageCount-1;j++) {
			if(mapPos<trigger[i].stage[j].trigger) continue;
			if(mapPos>trigger[i].stage[j+1].trigger) continue;
			// in the range.
			break;
		}
		if(j<trigger[i].stageCount-1) {
			int k;
			for(k=0;k<enemyCount;k++) {
				if(enemy[k].active==0) break;
			}
			if(k>=MAXENEMY) continue;
			//printf("Adding enemy %d from trigger %d\n",k,i);
			int eM=findEnemyModel(trigger[i].type);
			initActor(enemy+k,enemyModel+eM);
			if(k==enemyCount) enemyCount=k+1;
			enemy[k].trigger=trigger+i;
			trigger[i].actor=enemy+k;
			enemy[k].requestedMotion=MT_IDLE;
			enemy[k].health=trigger->health;
			if(enemy[k].health==0) enemy[k].enemyType=ET_DECORATION;
			//ASSERT: updateActor(enemy+k,elapsed,mapPos); must be done after this.
		}
	}
}

void updateTeather(struct Teather *teather,ScePspFVector3 *from,ScePspFVector3 *to)
{
	if(!teather->vert) {
		teather->vert=malloc(sizeof(struct Vertex3DCP)*130);
		teather->vertCount=128;
	}
	teather->from.x=from->x;
	teather->from.y=from->y;
	teather->from.z=from->z;
	teather->to.x=to->x;
	teather->to.y=to->y;
	teather->to.z=to->z;
}

void updateCollide(unsigned long elapsed)
{
	if(!lockOn) return;
	// Set up projection matricies
	int i;
	
	ScePspFMatrix4 proj;
	gumLoadIdentity(&proj);
	gumPerspective(&proj,45.0f,16.0f/9.0f,2.0f,1000.0f);
	ScePspFMatrix4 model;
	gumLoadIdentity(&model);
	fastPreProject(&proj,&view,&model);
	static s32 viewport[4]={0,0,480,272};
	ScePspFVector3 scr;
	ScePspFVector3 min={0,0,0},max={0,0,0};
	for(i=0;i<enemyCount;i++) {
		if(targeted>=8) break;
		if(enemy[i].targeted && enemy[i].teather.when+100>level.segment[level.activeSegment].elapsed) continue;	// Already targeted.
 		if(!enemy[i].active) continue;	// too late
 		if(enemy[i].enemyType==ET_DECORATION) continue;
 		if(enemy[i].targeted>=enemy[i].health) continue;
		scr.x=0;
		scr.y=0;

		ScePspFVector3 pos;
		int j;
		for(j=0;j<6;j++) {
			pos.x=enemy[i].pos.x+((j==0 || j==3)?enemy[i].minMax[j]:0);
			pos.y=enemy[i].pos.y+((j==1 || j==4)?enemy[i].minMax[j]:0);
			pos.z=enemy[i].pos.z+((j==2 || j==5)?enemy[i].minMax[j]:0);
			if(!fastProject(&pos,viewport,&scr)) break;
			if(j==0 || scr.x<min.x) min.x=scr.x;
			if(j==0 || scr.y<min.y) min.y=scr.y;
			if(j==0 || scr.x>max.x) max.x=scr.x;
			if(j==0 || scr.y>max.y) max.y=scr.y;
		}
		if(j<6) continue;
		// now see if the object is in the crop reticle.
		if(reticle.x+10<min.x || reticle.x-10>max.x || reticle.y+10<min.y || reticle.y-10>max.y) continue;
		playSfx(SFX_LOCKON);
		targeted++;
		enemy[i].targeted++;
		enemy[i].teather.when=level.segment[level.activeSegment].elapsed;
		enemy[i].teather.from=hero.pos;
		enemy[i].teather.to=enemy[i].pos;
		//printf("Targeted enemy %d\n",i);
	}
}

void updateChaos(unsigned long elapsed)
{
	if(chaosLeft==0) return;
	chaosLeft-=elapsed;
	if(chaosLeft<=0) {
		chaosLeft=0;
		int i;
		for(i=0;i<enemyCount;i++) {
			triggerExplosion(enemy+i);
		}
		targeted=0;
		return;
	}
	int i;
	for(i=0;i<enemyCount;i++) {
		if(!enemy[i].active) continue;
 		if(enemy[i].enemyType==ET_DECORATION) continue;
 		if(enemy[i].targeted>=enemy[i].health) continue;
		float dist=(hero.pos.x-enemy[i].pos.x)*(hero.pos.x-enemy[i].pos.x)+
		(hero.pos.y-enemy[i].pos.y)*(hero.pos.y-enemy[i].pos.y)+
		(hero.pos.z-enemy[i].pos.z)*(hero.pos.z-enemy[i].pos.z);
		if(dist>250*250.0f*(chaosMax-chaosLeft)/chaosMax) continue;
		playSfx(SFX_LOCKON);
		enemy[i].targeted++;
		enemy[i].teather.when=level.segment[level.activeSegment].elapsed;
		enemy[i].teather.from=hero.pos;
		enemy[i].teather.to=enemy[i].pos;
		printf("Chaos targets enemy %d\n",i);		
	}
}

void updateCamera(unsigned long elapsed)
{
	// Update camera
	fromTarget.x=0;
	fromTarget.y=85;
	fromTarget.z=0;
	int timeleft;
	switch(level.segment[level.activeSegment].cameraAngle) {
	case 1:
		fromTarget.x+=-155;
		if(level.segment[level.activeSegment].elapsed<3000) {
			// rotate into place.
			float per=level.segment[level.activeSegment].elapsed/3000.0f;
			float cent=1.0f-per;
			fromTarget.z=-155*cent;
			fromTarget.x*=per;
		}
		timeleft=level.segment[level.activeSegment].duration-level.segment[level.activeSegment].elapsed;
		if(timeleft<5000) {
			// rotate into place.
			float per=timeleft/3000.0f;
			if(per>1) per=1.0f;
			float cent=1.0f-per;
			fromTarget.z=-155*cent;
			fromTarget.x*=per;
		}	
		break; // left
	case 2:
		fromTarget.x+=155;
		if(level.segment[level.activeSegment].elapsed<3000) {
			// rotate into place.
			float per=level.segment[level.activeSegment].elapsed/3000.0f;
			float cent=1.0f-per;
			fromTarget.z=-155*cent;
			fromTarget.x*=per;
		}
		timeleft=level.segment[level.activeSegment].duration-level.segment[level.activeSegment].elapsed;
		if(timeleft<5000) {
			// rotate into place.
			float per=timeleft/3000.0f;
			if(per>1) per=1;
			float cent=1.0f-per;
			fromTarget.z=-155*cent;
			fromTarget.x*=per;
		}	
		break; // right
	default:
		fromTarget.z-=155;
		break; // forward
	}
	// Now rotate around y axis then x axis
	float rx,ry,rz;
	float ax,ay;
	ay=-(reticle.x-240)/640.0f;
	ax=-(reticle.y-136)/640.0f;
	if(lookAround) {
		ay*=4;
		ax*=4;
	}
	rx=fromTarget.x*cosf(ay)+fromTarget.z*sinf(ay);
	ry=fromTarget.y;
	rz=-fromTarget.x*sinf(ay)+fromTarget.z*cosf(ay);
	fromTarget.x=rx;
	fromTarget.y=ry*cosf(ax)+rz*sinf(ax);
	fromTarget.z=-ry*sinf(ax)+rz*cosf(ax);
	fromTarget.x+=to.x;
	fromTarget.y+=to.y;
	fromTarget.z+=to.z;
/*	from.x=from.x*0.95f+fromTarget.x*0.05f;
	from.y=from.y*0.95f+fromTarget.y*0.05f;
	from.z=from.z*0.95f+fromTarget.z*0.05f; */
	from.x=fromTarget.x;
	from.y=fromTarget.y;
	from.z=fromTarget.z;
}

void update(unsigned long elapsed)
{
	if(gameModeTimer>0) {
		gameModeTimer--;
		if(gameModeTimer==0) {
			if(newMode==GAMEMODE && gameMode!=PAUSEMODE) {
				newGame();
				gameModeTimer=0;
				gameMode=newMode;
			}
			if(newMode==SUMMARYMODE) {
				hero.active=1;	// Restore the player after death.
			}
			gameMode=newMode;
			pauseMode=newPauseMode;
			menuItem=0;
			menuTimer=0;
			alertMessage[0]=0;
		}			
	}
	float dx=0,dy=0;
	if(currentDir>0) {
		int bit;
		for(bit=0;bit<4;bit++) {
			if(currentDir&(1<<bit)) {
				dx+=dir[bit].x;
				dy+=dir[bit].y;
			}
		}
	}
	dx+=px;
	dy+=py;
	if(options.invert) dy=-dy;
	if(dx!=0 || dy!=0) {
		float after;
		after=reticle.x+elapsed*options.speed*dx;
		if(after>0.0f && after<480.0f) reticle.x=after;
		after=reticle.y+elapsed*options.speed*dy;
		if(after>0.0f && after<272.0f) reticle.y=after;
	}

	// Do triggers for creating enemies and so on.
	int mapPos=level.segment[level.activeSegment].elapsed;
	int i;
	updateTrigger(mapPos);

	// Update the actors (hero, and enemies).
	updateActor(&hero,elapsed,mapPos);	
	if(invulnerableTimer>0) {
		invulnerableTimer-=elapsed;
		if(invulnerableTimer<0) invulnerableTimer=0;
	}
	for(i=0;i<enemyCount;i++) {
		updateActor(enemy+i,elapsed,mapPos);
		if(enemy[i].targeted) updateTeather(&enemy[i].teather,&enemy[i].pos,&hero.pos);
		if(enemy[i].active && enemy[i].enemyType==ET_ACTIVE) {
			float dist=(hero.pos.x-enemy[i].pos.x)*(hero.pos.x-enemy[i].pos.x)+
			(hero.pos.y-enemy[i].pos.y)*(hero.pos.y-enemy[i].pos.y)+
			(hero.pos.z-enemy[i].pos.z)*(hero.pos.z-enemy[i].pos.z);
			float heroSize=(hero.minMax[3]-hero.minMax[0])/2;
			heroSize=heroSize*heroSize;
			float enemySize=(enemy[i].minMax[3]-enemy[i].minMax[0])/2;
			enemySize=enemySize*enemySize;
			if(dist<heroSize+enemySize) {
				if(hero.health>0 && invulnerableTimer==0 && gameMode==GAMEMODE && enemy[i].targeted==0) {
					hero.health--;
					if(hero.health>0) hero.model=heroModel+(hero.health-1);
					dropMessage(BT_NOTHING,&hero);	// level down
					invulnerableTimer=2000;
				}
				if(hero.health==0 && invulnerableTimer==0 && gameMode==GAMEMODE && enemy[i].targeted==0) {
					hero.targeted=1;
					hero.health=1;
					triggerExplosion(&hero);
					playSfx(SFX_LOSE);
					setMode(SUMMARYMODE);
				} else {
					enemy[i].targeted=enemy[i].health;
					triggerExplosion(enemy+i);
				}
			}
		}
	}

	// Do any pending level animation.
	if(gameMode!=PAUSEMODE) {
		level.elapsed+=elapsed;
		level.segment[level.activeSegment].duration=level.duration/level.segmentCount;
	}
//	if(level.elapsed>level.duration/level.segmentCount*(level.activeSegment+1)) {
	if(level.segment[level.activeSegment].elapsed>level.segment[level.activeSegment].duration) {
		if(level.segment[level.activeSegment].isBoss && gameMode==GAMEMODE) {
			if(level.segment[level.activeSegment].bossHealth<0 || level.bossHealth<=0) {
				level.activeSegment++;
			}
		} else {
			level.activeSegment++;
		}
		for(i=0;i<triggerCount;i++) trigger[i].actor=0;
		for(i=0;i<enemyCount;i++) {
			if(gameMode==GAMEMODE && enemy[i].active) {
				if(enemy[i].enemyType==ET_ACTIVE) enemyMiss++;
				else if(enemy[i].enemyType==ET_CHAOSUP || 
				enemy[i].enemyType==ET_EVOUP) supportMiss++;
			}
			if(enemy[i].trigger) enemy[i].trigger->actor=0;
			enemy[i].active=0;
		}
		enemyCount=0;
		targeted=0;
		hero.requestedMotion=MT_IDLE;
		if(level.activeSegment>=level.segmentCount) {
			level.activeSegment=0;
			level.elapsed=0;
			if(gameMode==GAMEMODE) {
				playSfx(SFX_WIN);
				setMode(SUMMARYMODE);
			}
		}
		if(level.segment[level.activeSegment].isBoss) {
			int bossHealth=(level.bossHealth-1)%(level.bossHealthMax/4);
			if(level.bossHealth<=0) bossHealth=-1;
			level.segment[level.activeSegment].bossHealth=bossHealth;
		}
		for(i=0;i<explosionCount;i++) {
			explosion[i].active=0;
		}
		level.segment[level.activeSegment].elapsed=0;
		genTerrainNoPoly();
	} else {
		if(gameMode!=PAUSEMODE) level.segment[level.activeSegment].elapsed+=elapsed;
	}
	
	int tile=256/grid;
	float z=level.segment[level.activeSegment].elapsed/(float)level.segment[level.activeSegment].duration*grid*level.segment[level.activeSegment].length*tile;
	to.z=z;
	hero.pos.z=z;
	
	updateCamera(elapsed);
	updateChaos(elapsed);
	
	// Time for collisions if we can
	updateCollide(elapsed);

	// update explosion
	for(i=0;i<explosionCount;i++) {
		if(explosion[i].active) {
			explosion[i].elapsed+=(int)elapsed;
			if(explosion[i].elapsed>explosion[i].duration) explosion[i].active=0;
		}
	}
}

void handleChaos()
{
	if(chaosUp<1) return;
	if(chaosLeft>0) return;
	chaosLeft=chaosMax;
	chaosUp--;
	playSfx(SFX_CHAOS);
}

void dropItem(enum BonusType bt,struct Actor *actor)
{
	const char *enemyModelName[]={ "","powerup/bonuslevelup.md2","powerup/bonuschaosup.md2","powerup/bonusevoup.md2","",};
	int i,k;
	for(k=0;k<enemyCount;k++) {
		if(enemy[k].active==0) break;
	}
	if(k==enemyCount) enemyCount=k+1;
printf("Adding bonus %d from enemy\n",k);
	int eM=findEnemyModel(enemyModelName[bt]);
	initActor(enemy+k,enemyModel+eM);
	for(i=0;i<bonusTriggerCount;i++) {
		if(bonusTrigger[i].actor==0) break;
	}
	if(i==MAXBONUSTRIGGER) {
		enemy[k].active=0;
		printf("Deactivating the bonus.  Out of bonuses!\n");
		return;	// an error to be sure.
	}
	bonusTrigger[i].actor=&enemy[k];
	strcpy(bonusTrigger[i].type,enemyModelName[bt]);
	int elapsed=level.segment[level.activeSegment].elapsed;
	int duration=level.segment[level.activeSegment].duration;
	float length=level.segment[level.activeSegment].length;
	bonusTrigger[i].health=1;
	bonusTrigger[i].points=2;
	bonusTrigger[i].segment=level.activeSegment;
	bonusTrigger[i].stageCount=3;
	bonusTrigger[i].stage[0].trigger=elapsed;
	bonusTrigger[i].stage[0].pos=actor->pos;
	bonusTrigger[i].stage[0].orient=actor->orient;
	bonusTrigger[i].stage[1].trigger=elapsed+2000;
	bonusTrigger[i].stage[1].pos.x=-50;
	bonusTrigger[i].stage[1].pos.y=actor->pos.y;
	bonusTrigger[i].stage[1].pos.z=(elapsed+3000)*length*256/duration;
	bonusTrigger[i].stage[1].orient=actor->orient;	
	bonusTrigger[i].stage[2].trigger=elapsed+8000;
	bonusTrigger[i].stage[2].pos.x=(actor->pos.x<0)?-50:50;
	bonusTrigger[i].stage[2].pos.y=(actor->pos.y<-35)?-50:50;
	bonusTrigger[i].stage[2].pos.z=(elapsed+9000)*length*256/duration;
	bonusTrigger[i].stage[2].orient=actor->orient;	
	if(bonusTriggerCount==i) bonusTriggerCount++;
	enemy[k].trigger=bonusTrigger+i;
	enemy[k].requestedMotion=MT_IDLE;
	enemy[k].health=bt==BT_EVOUP?1:8;
	switch(bt) {
	case BT_EVOUP: enemy[k].enemyType=ET_EVOUP; break;
	case BT_CHAOSUP: enemy[k].enemyType=ET_CHAOSUP; break;
	case BT_LEVELUP: enemy[k].enemyType=ET_LEVELUP; break;
	default: enemy[k].enemyType=ET_ACTIVE; break;
	}
	updateActor(enemy+k,0,elapsed);
}

void dropMessage(enum BonusType bt,struct Actor *actor)
{
	const char *enemyModelName[]={ "powerup/leveldown.md2","powerup/levelup.md2","powerup/chaosup.md2","powerup/evoup.md2","",};
	int i=0,k;
	for(k=0;k<enemyCount;k++) {
		if(enemy[k].active==0) break;
	}
	if(k==enemyCount) enemyCount=k+1;
	int eM=findEnemyModel(enemyModelName[bt]);
printf("Adding message %d from enemy %d (model %d)\n",k,i,eM);
	initActor(enemy+k,enemyModel+eM);
	for(i=0;i<bonusTriggerCount;i++) {
		if(bonusTrigger[i].actor==0) break;
	}
	if(i>=MAXBONUSTRIGGER) {
		enemy[k].active=0;
		printf("Deactivating the messages.  Out of messages!\n");
		return;	// an error to be sure.
	}
	bonusTrigger[i].actor=&enemy[k];
	strcpy(bonusTrigger[i].type,enemyModelName[bt]);
	int elapsed=level.segment[level.activeSegment].elapsed;
	int duration=level.segment[level.activeSegment].duration;
	float length=level.segment[level.activeSegment].length;
	bonusTrigger[i].enemyType=ET_DECORATION;
	bonusTrigger[i].health=0;
	bonusTrigger[i].points=3;
	bonusTrigger[i].segment=level.activeSegment;
	bonusTrigger[i].stageCount=3;
	bonusTrigger[i].stage[0].trigger=elapsed;
	bonusTrigger[i].stage[0].pos=actor->pos;
	bonusTrigger[i].stage[0].orient.x=0;
	bonusTrigger[i].stage[0].orient.y=0;
	bonusTrigger[i].stage[0].orient.z=0;
	bonusTrigger[i].stage[1].trigger=elapsed+2000;
	bonusTrigger[i].stage[1].pos.x=0;
	bonusTrigger[i].stage[1].pos.y=-35;
	bonusTrigger[i].stage[1].pos.z=(elapsed+2000)*length*256/duration;
	bonusTrigger[i].stage[1].orient.x=0;
	bonusTrigger[i].stage[1].orient.y=GU_PI*2;
	bonusTrigger[i].stage[1].orient.z=0;
	bonusTrigger[i].stage[2].trigger=elapsed+4000;
	bonusTrigger[i].stage[2].pos.x=-35;
	bonusTrigger[i].stage[2].pos.y=-256;
	bonusTrigger[i].stage[2].pos.z=(elapsed+4000)*length*256/duration;
	bonusTrigger[i].stage[2].orient.x=0;
	bonusTrigger[i].stage[2].orient.y=0;
	bonusTrigger[i].stage[2].orient.z=0;
	if(bonusTriggerCount==i) bonusTriggerCount++;
	enemy[k].enemyType=ET_DECORATION;
	enemy[k].trigger=bonusTrigger+i;
	enemy[k].requestedMotion=MT_IDLE;
	enemy[k].health=bt==BT_EVOUP?1:8;
	updateActor(enemy+k,0,elapsed);
}

void triggerBossExplosion(int damage);
void triggerExplosion(struct Actor *enemy)
{
	if(!enemy->targeted) return;
	enemy->health-=enemy->targeted;
	if(enemy->enemyType==ET_ACTIVE) triggerBossExplosion(enemy->targeted);
	// queue up sounds.
	
	// did we kill it?
	enemy->targeted=0;
	if(enemy->health>0) {
		playSfx(SFX_EXPLODE);
		// should emit dammage rings
		return;
	}
	// we beat it:
	if(gameMode==GAMEMODE && enemy->enemyType==ET_ACTIVE) {
		enemyHit++;
		evoKills++;
		if(evoKills>=level.evoDropRate) {
			evoKills=0;
			dropItem(BT_EVOUP,enemy);
		}
		chaosKills++;
		if(chaosKills>=level.chaosDropRate) {
			chaosKills=0;
			dropItem(BT_CHAOSUP,enemy);
		}
		if(enemy->trigger && enemy->trigger->bonusType!=BT_NOTHING) {
			dropItem(enemy->trigger->bonusType,enemy);
		}
	}
	float bonusMult=1;
	if(targeted>1) bonusMult=bonusMult+targeted*0.25f;
	if(enemy->trigger) {
		switch(enemy->trigger->points) {
		case 0: gameScore+=100*bonusMult; break;
		case 1: gameScore+=250*bonusMult; break;
		case 2: gameScore+=1000*bonusMult; break;
		}
	}
	scoreBonus=bonusMult*100-100;
	int j;
	for(j=0;j<explosionCount;j++) {
		if(explosion[j].active==0) break;
	}
	if(j==explosionCount) j=0;
	initExplosion(explosion+j,&enemy->pos,&enemy->orient);
	switch(enemy->enemyType) {
	case ET_EVOUP:
		playSfx(SFX_POWERUP);
		evoUp++;
		supportHit++;
		if(evoUp<8) {
			dropMessage(BT_EVOUP,enemy);
		} else if(hero.health<5) {
			evoUp-=8;
			hero.health++;
			hero.model=heroModel+(hero.health-1);
			dropMessage(BT_LEVELUP,enemy);
		}
		break;
	case ET_CHAOSUP:
		playSfx(SFX_POWERUP);
		chaosUp++;
		dropMessage(BT_CHAOSUP,enemy);
		supportHit++;
		break;
	case ET_LEVELUP:
		playSfx(SFX_POWERUP);
		dropMessage(BT_LEVELUP,enemy);
		supportHit++;
		break;
	default:
		playSfx(SFX_EXPLODE);
		break;
	}
	if(enemy->trigger && enemy->trigger->bonusType!=BT_NOTHING) {
		dropItem(enemy->trigger->bonusType,enemy);
	}
	enemy->active=0;
}

void triggerBossExplosion(int damage)
{
	if(!level.segment[level.activeSegment].isBoss) return;
	int mod=level.bossHealthMax/4;
	int bossHealth=(level.bossHealth-1)%mod;
	if(level.bossHealth<=0) return;
	level.bossHealth-=damage;
	level.segment[level.activeSegment].bossHealth-=damage;
	if(level.bossHealth<0) {
		level.bossHealth=0;
	}
	if(level.segment[level.activeSegment].bossHealth>=0) return;
	if(level.segment[level.activeSegment].bossHealth+damage<0) return;
	
	// queue up sounds.	
	level.bossTimer=level.elapsed;
	playSfx(SFX_BOSSDAMAGE);
	
	int i;
	for(i=0;i<9;i++) {
		int j;
		for(j=0;j<explosionCount;j++) {
			if(explosion[j].active==0) break;
		}
		if(j==explosionCount) j=0;
		ScePspFVector3 pos={0,0,0},orient={0,0,0};
		pos.x=hero.pos.x-48*i+48*5;
		pos.y=hero.pos.y-120+30;
		pos.z=hero.pos.z+256+50;
		initExplosion(explosion+j,&pos,&orient);
		explosion[j].firstColor=GU_RGBA(128,192,192,255);
		explosion[j].lastColor=GU_RGBA(0,128,192,255);
	}
}

int handleJoyGame(enum Buttons button,int up)
{	
	int i;
	if(up==1) {
		switch(button) {
		case BT_DOWN:
			currentDir&= ~(1<<2);
			break;
		case BT_UP:
			currentDir&= ~(1<<0);
			break;
		case BT_LEFT:
			currentDir&= ~(1<<3);
			break;
		case BT_RIGHT:
			currentDir&= ~(1<<1);
			break;
		case BT_CROSS:
			lockOn=0;
			for(i=0;i<enemyCount;i++) {
				triggerExplosion(enemy+i);
			}
			targeted=0;
			break;
		case BT_LTRIGGER:
			lookAround=0;
			break;
		default:
			break;
		}
		return 0;
	}
	switch(button) {
	case BT_DOWN:
		currentDir|=1<<2;
		break;
	case BT_UP:
		currentDir|=1<<0;
		break;
	case BT_LEFT:
		currentDir|=1<<3;
		break;
	case BT_RIGHT:
		currentDir|=1<<1;
		break;
	case BT_SQUARE:
		handleChaos();
		break;
	case BT_CROSS:
		lockOn=1;
		playSfx(SFX_TARGET);
		break;
	case BT_LTRIGGER:
		lookAround=1;
		break;
	case BT_RTRIGGER:
		showScore=1-showScore;
		break;
	case BT_TRIANGLE:
		showFps=1-showFps;
		break;
	case BT_CIRCLE:
		break;
	default:
		break;
	}
	if(button==BT_START) setMode(PAUSEMODE);
	return 0;
}

int handleAnalog(int lx,int ly)
{
	px=0;
	py=0;
	if(ly<64) {
		py=(ly-64)/64.0f;
	} else if(ly>192) {
		py=(ly-192)/64.0f;
	}
	if(lx<64) {
		px=(lx-64)/64.0f;
	} else if(lx>192) {
		px=(lx-192)/64.0f;
	}
	//printf("lxy=(%d,%d); pxy=(%f,%f)\n",lx,ly,px,py);
	return 0;
}
