#include<math.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<malloc.h>

#ifdef _PSP
#include<pspctrl.h>
#include<pspgu.h>
#include<pspgum.h>
#endif
#include "main.h"

struct Material {
	char name[64];
	unsigned long color;
	Image *image;
} material[256];
int nextMaterial;

struct Item {
    char name[64];
    float matrix[16];
    int vertCount;
    struct Image *image;
    struct Vertex3DCNP *vert;
    struct Vertex3DTNP *vertImage;
    int polyCount;
    unsigned short *poly;
};
#define MAXITEMS 32
struct Item item[MAXITEMS];
int nextItem;

void loadItemImage(char *fname)
{
	char path[256];
	struct Vertex3DTP *vertImage=0;
	nextMaterial=0;
	int fCount=0;
	
	printf("read texture for '%s'\n",fname);
	sprintf(path,"data/%s.png",fname);
	item[nextItem].image=loadPng(path);
	if(!item[nextItem].image) {
		sprintf(path,"models/boss/%s.png",fname);
		item[nextItem].image=loadPng(path);
	}
	//swizzle_fast(item[nextItem].image);
	swizzleFast(item[nextItem].image);
	if(!item[nextItem].image) printf("*** Could not load texture '%s'\n",path);
	
	printf("read materials for '%s'\n",fname);
	sprintf(path,"models/boss/%s.mtl",fname);
//	gzFile file=gzopen(path,"r");
	FILE *file=fopen(path,"r");
		strcpy(item[nextItem].name,fname);
	if(file) {
		// Read in the materials.
		char line[256];
		line[255]=0;
		int mat=nextMaterial;

//		while( gzgets(file,line,255) ) {
		while( fgets(line,255,file) ) {
			char cmd[64];

			sscanf(line,"%s",cmd);
			if(strchr(line,'\r')) strchr(line,'\r')[0]=0;
			if(strchr(line,'\n')) strchr(line,'\n')[0]=0;

			if(strcmp(cmd,"newmtl")==0 ) {
				mat=nextMaterial;
				nextMaterial++;
				strcpy(material[mat].name,line+7);
			} else if( strcmp(cmd,"Kd")==0) {
				float rf=0,gf=0,bf=0;
				unsigned long r,g,b;
				sscanf(line,"Kd%f%f%f",&rf,&gf,&bf);
				r=rf*255;
				g=gf*255;
				b=bf*255;
				material[mat].color=(r)|(g<<8)|(b<<16)|(255<<24);
				//printf("%s is %d/%d/%d\n",material[mat].name,(int)r,(int)g,(int)b);
			}

		}
		fclose(file);
//		gzclose(file);
		printf("read %d materials for %s\n",nextMaterial,fname);
	}

	printf("read obj for '%s'\n",fname);
	sprintf(path,"models/boss/%s.obj",fname);
	//sprintf(path,"data/intro/%s.obj.gz",fname);
//	file=gzopen(path,"r");
	file=fopen(path,"r");

	if(!file) printf("Couldn't find %s\n",path);
	if(file) {
		// Count the verts.
		char line[256];
		line[255]=0;
		int i;
		int v=0;
		int vt=0;
		int f=0;

//		while( gzgets(file,line,255) ) {
		while( fgets(line,255,file) ) {
//printf("Pass 1: %s",line);			
			switch(line[0]) {
			case 'v':
				if(line[1]==' ') v++;
				if(line[1]=='t') vt++;
				break;
			case 'f':
				f+=3;
				break;
			}
		}
		printf("Found %d image verts, %d texture verts, and %d faces.\n",v,vt,f);
		strcpy(item[nextItem].name,fname);
		item[nextItem].polyCount=f;
		item[nextItem].poly=calloc(sizeof(short),f);
		item[nextItem].vertCount=v;
		vertImage=calloc(sizeof(struct Vertex3DTP),v<vt?vt:v);
		item[nextItem].vertImage=calloc(sizeof(struct Vertex3DTNP),f);
		fCount=f;
		//printf("item[nextItem].vert=%08x\n",(int)item[nextItem].vert);
		for(i=0;i<16;i++) {
			item[nextItem].matrix[i]=(i%4)==(i/4)?1:0;
		}
		//gzclose(file);
		fclose(file);
		file=0;
	}
//	file=gzopen(path,"r");
	file=fopen(path,"r");
	if(file) {
		// actual verts
		char line[256];
		line[255]=0;
		int v=0;
		int vt=0;
		//int gv=0;
		int f=0;
		int m=0;
		float min[3]={0,0,0},max[3]={0,0,0};

//		while( gzgets(file,line,255) ) {
		while( fgets(line,255,file)) {
			float tu,tv,tw;
			float x,y,z;
			int f1,f2,f3;	// face vertex
			int n1,n2,n3;	// normal
			int t1,t2,t3;	// texture
			char name[64];

			if(sscanf(line,"v%f%f%f",&x,&y,&z)==3) {
				if(x<min[0]) min[0]=x;
				if(y<min[1]) min[1]=y;
				if(z<min[2]) min[2]=z;
				if(x>max[0]) max[0]=x;
				if(y>max[1]) max[1]=y;
				if(z>max[2]) max[2]=z;
				vertImage[v].x=x;
				vertImage[v].y=y;
				vertImage[v].z=z;
				v++;
				//if(v>item[nextItem].vertCount) {
					//printf("At indexed vert %d of %d\n",v,item[nextItem].vertCount);
				//}
			} else if(sscanf(line,"vt%f%f%f",&tu,&tv,&tw)==3) {
				vertImage[vt].u=tu;
				vertImage[vt].v=1.0f-tv;
				vt++;
				//if(v>item[nextItem].vertCount) {
					//printf("At indexed vert %d of %d\n",v,item[nextItem].vertCount);
				//}
			} else if(line[0]=='g') {
				//gv=v;
				//printf("Got group at %d (%d)\n",v,gv);
			} else if(sscanf(line,"usemtl %s",name)>0) {
				for(m=0;m<nextMaterial;m++) {
					if(strcmp(name,material[m].name)==0) break;
				}
				if(m==nextMaterial) {
					printf("Material error.  Couldn't find %s\n",name);
					m=0;	// bad file
				}
				//printf("Applying material %s (%d) at %d-%d\n",name,m,gv,v);
				//while(gv<v) {
				//	vert[gv++].color=material[m].color;
				//}
			} else if(sscanf(line,"f%d/%d/%d%d/%d/%d%d/%d/%d",&f1,&t1,&n1,&f2,&t2,&n2,&f3,&t3,&n3)==9||
				sscanf(line,"f%d/%d%d/%d%d/%d",&f1,&t1,&f2,&t2,&f3,&t3)==6) {
				if(f1<0) f1=v-f1+1;
				if(f2<0) f2=v-f2+1;
				if(f3<0) f3=v-f3+1;
				if(t1<0) t1=vt-t1+1;
				if(t2<0) t2=vt-t2+1;
				if(t3<0) t3=vt-t3+1;
				item[nextItem].vertImage[f].x=vertImage[f1-1].x;
				item[nextItem].vertImage[f].y=vertImage[f1-1].y;
				item[nextItem].vertImage[f].z=vertImage[f1-1].z;
				item[nextItem].vertImage[f].u=vertImage[t1-1].u;
				item[nextItem].vertImage[f].v=vertImage[t1-1].v;
				item[nextItem].poly[f++]=f1-1;
				item[nextItem].vertImage[f].x=vertImage[f2-1].x;
				item[nextItem].vertImage[f].y=vertImage[f2-1].y;
				item[nextItem].vertImage[f].z=vertImage[f2-1].z;
				item[nextItem].vertImage[f].u=vertImage[t2-1].u;
				item[nextItem].vertImage[f].v=vertImage[t2-1].v;
				item[nextItem].poly[f++]=f2-1;
				item[nextItem].vertImage[f].x=vertImage[f3-1].x;
				item[nextItem].vertImage[f].y=vertImage[f3-1].y;
				item[nextItem].vertImage[f].z=vertImage[f3-1].z;
				item[nextItem].vertImage[f].u=vertImage[t3-1].u;
				item[nextItem].vertImage[f].v=vertImage[t3-1].v;
				item[nextItem].poly[f++]=f3-1;
				float v1[3],v2[3];
				float nx,ny,nz;
				v1[0]=item[nextItem].vertImage[f-2].x-item[nextItem].vertImage[f-3].x;
				v1[1]=item[nextItem].vertImage[f-2].y-item[nextItem].vertImage[f-3].y;
				v1[2]=item[nextItem].vertImage[f-2].z-item[nextItem].vertImage[f-3].z;
				v2[0]=item[nextItem].vertImage[f-1].x-item[nextItem].vertImage[f-3].x;
				v2[1]=item[nextItem].vertImage[f-1].y-item[nextItem].vertImage[f-3].y;
				v2[2]=item[nextItem].vertImage[f-1].z-item[nextItem].vertImage[f-3].z;
				// A x B = <Ay*Bz - Az*By, Az*Bx - Ax*Bz, Ax*By - Ay*Bx>
				nx=v1[1]*v2[2]-v1[2]*v2[1];
				ny=v1[2]*v2[0]-v1[0]*v2[2];
				nz=v1[0]*v2[1]-v1[1]*v2[0];
				float nw=sqrtf(nx*nx+ny*ny+nz*nz);
				if(nw==0) nw=1;
				nx=nx/nw;
				ny=ny/nw;
				nz=nz/nw;
				item[nextItem].vertImage[f-3].nx=nx;
				item[nextItem].vertImage[f-3].ny=ny;
				item[nextItem].vertImage[f-3].nz=nz;
				item[nextItem].vertImage[f-2].nx=nx;
				item[nextItem].vertImage[f-2].ny=ny;
				item[nextItem].vertImage[f-2].nz=nz;
				item[nextItem].vertImage[f-1].nx=nx;
				item[nextItem].vertImage[f-1].ny=ny;
				item[nextItem].vertImage[f-1].nz=nz;
			}
		}
		fclose(file);
//		gzclose(file);
		printf("Min: %.2f,%.2f,%.2f; Max: %.2f,%.2f,%.2f\n",min[0],min[1],min[2],max[0],max[1],max[2]);
		float centre[3];
		centre[0]=(max[0]+min[0])/2;
		centre[1]=(max[1]+min[1])/2;
		centre[2]=(max[2]+min[2])/2;
/*
		int i;
		for(i=0;i<fCount;i++) {
			item[nextItem].vertImage[i].x-=centre[0];
			item[nextItem].vertImage[i].y-=centre[1];
			item[nextItem].vertImage[i].z-=centre[2];
		}
*/
		free(vertImage);
		vertImage=0;
		free(item[nextItem].poly);
		item[nextItem].poly=0;
		nextItem++;
	}
}

void loadItems()
{
	if(nextItem>0) return;	// Only once.
	loadItemImage("ship-full");
	loadItemImage("ship-75");
	loadItemImage("ship-50");
	loadItemImage("ship-25");
}

void freeItems()
{
	int i;
	for(i=0;i<nextItem;i++) {
		if(item[i].image) freeImage(item[i].image);
		item[i].image=0;
		if(item[i].vertImage) free(item[i].vertImage);
		item[i].vert=0;
		if(item[i].vert) free(item[i].vert);
		item[i].vert=0;
		if(item[i].poly) free(item[i].poly);
		item[i].poly=0;
	}
	nextItem=0;
}

int getItemCount()
{
	return nextItem;
}

void renderItem(int i)
{
	if(i>=nextItem || i<0) return;
#ifdef _PSP
//	printf("Rendering item %d\n",i);
	if(item[i].image) {
		Image *source=item[i].image;
		sceGuTexMode(GU_PSM_8888, 0, 0, source->isSwizzled);
		sceGuTexImage(0, source->textureWidth, source->textureHeight, source->textureWidth, source->data);
		sceGuTexScale(1.0f,1.0f);
		sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
		sceGuEnable(GU_LIGHTING);
		//printf("showing image %08x\n",(int)source);
	}
	//printf("Drawing item %d\n",i);
	sceGumMatrixMode(GU_MODEL);
	sceGumPushMatrix();
	sceGumMultMatrix((ScePspFMatrix4 *)item[i].matrix);
	int j=0;
	if(item[i].image) {
		sceGuEnable(GU_TEXTURE_2D);
		sceGuColor(0xffffffff);
	}
	sceGuFrontFace(GU_CCW);
	while(j<item[i].polyCount) {
		int count=30720;
		if(j+count>item[i].polyCount) count=item[i].polyCount-j;
//		printf("Drawing %d polys\n",count);
		//sceGumDrawArray(GU_TRIANGLES,GU_VERTEX_32BITF|GU_NORMAL_32BITF|GU_COLOR_8888|GU_INDEX_16BIT|GU_TRANSFORM_3D,count,item[i].poly+j,item[i].vert);
		if(item[i].vert)
			sceGumDrawArray(GU_TRIANGLES,GU_VERTEX_32BITF|GU_NORMAL_32BITF|GU_COLOR_8888|GU_TRANSFORM_3D,count,0,item[i].vert+j);
		if(item[i].vertImage) {
			sceGumDrawArray(GU_TRIANGLES,GU_VERTEX_32BITF|GU_NORMAL_32BITF|GU_TEXTURE_32BITF|GU_TRANSFORM_3D,count,0,item[i].vertImage+j);
		}
		j=j+count;
	}
	sceGuFrontFace(GU_CW);
	if(item[i].image) {		
		sceGuDisable(GU_LIGHTING);
		sceGuTexScale(1,1);
	}
	sceGumPopMatrix();
#endif
}

void setItemPos(int i,float x,float y,float z)
{
	if(i<0 || i>=nextItem) return;
	item[i].matrix[12]=x;
	item[i].matrix[13]=y;
	item[i].matrix[14]=z;
	item[i].matrix[0]=6;
	item[i].matrix[5]=6;
	item[i].matrix[10]=6;
}
