#include <psptypes.h>
#include <psphttp.h>

#include "main.h"

#if 1
#include <curl/curl.h>
#include <curl/easy.h>
CURL *curl;
	
void httpInit()
{
	curl=curl_easy_init();
}

void httpTerm()
{
	curl_easy_cleanup(curl);
}

struct CurlCallback {
	char *response;
	int responseUsed;
	int responseLength;
};

size_t curlWriteCallback(void *ptr,size_t size,size_t nmemb,void *data)
{
	size_t realsize=size*nmemb;
	struct CurlCallback *cb=(struct CurlCallback *)data;
	char *s=(char *)ptr;
	while(realsize>0 && cb->responseUsed<cb->responseLength) {
		cb->response[cb->responseUsed++]=*s;
		s++;
		realsize--;
	}
	return nmemb;
}

int post(const char *url,const char *body,int bodyLength,char *response,int *responseLength)
{
	CURLcode res;
	curl_easy_setopt(curl,CURLOPT_URL,url);
	if(body) {
		curl_easy_setopt(curl, CURLOPT_POSTFIELDS, body);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, bodyLength);
	}
	struct CurlCallback cb={response,0,*responseLength};
	if(response) {
		curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curlWriteCallback);
		curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&cb);
	}
	res=curl_easy_perform(curl);
	int status=0;
	curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE,&status);
	if(response) {
		printf("Result:\n%s\n",response);
		responseLength[0]=cb.responseUsed;
	} else {
		responseLength[0]=0;
		return status;
	}
	return responseLength;
}

#else
extern int httpTemplate;
void httpInit()
{
	sceHttpInit(20000);
    httpTemplate=sceHttpCreateTemplate("Resonate/0.50 libhttp/1.0.0",1,0);
}

void httpTerm()
{
	sceHttpEnd();
}

int post(const char *url,const char *body,int bodyLength,char *response,int *responseLength)
{
	int httpConnection;
	int httpRequest;
	int result;
	printf("Creating connection...\n");
	result=sceHttpSetResolveTimeOut(httpTemplate, 3000000);
	if(result<0) printf("ResolveTimeOut error %08x\n",result);
	result=sceHttpSetRecvTimeOut(httpTemplate, 60000000);
	if(result<0) printf("RecvTimeOut error %08x\n",result);
	result=sceHttpSetSendTimeOut(httpTemplate, 60000000);
	if(result<0) printf("SendTimeOut error %08x\n",result);

	httpConnection=sceHttpCreateConnectionWithURL(httpTemplate,url,0);
	if(httpConnection<0) {
		strcpy(alertMessage,"Can't connect to server!");
		*responseLength=0;
		return httpConnection;
	}
	httpRequest=sceHttpCreateRequestWithURL(httpConnection,bodyLength==0?PSP_HTTP_METHOD_GET:PSP_HTTP_METHOD_POST,(char *)url,bodyLength);
	if(httpRequest<0) {
		strcpy(alertMessage,"Can't make request to server!");
		*responseLength=0;
		sceHttpDeleteConnection(httpConnection);
		return httpRequest;
	}
	result=sceHttpSendRequest(httpRequest,(char *)body,bodyLength);
	if(result<0) {
		strcpy(alertMessage,"Can't send online request.");
		*responseLength=0;
		sceHttpDeleteRequest(httpRequest);
		sceHttpDeleteConnection(httpConnection);
		return result;
	}
	int statusCode=0;
	result=sceHttpGetStatusCode(httpRequest,&statusCode);
	if(result<0) {
		strcpy(alertMessage,"Bad status code while talking online");
		sceHttpDeleteRequest(httpRequest);
		sceHttpDeleteConnection(httpConnection);
		return result;
	}
	printf("Got status code %d\n",statusCode);
	if(statusCode==200) {
		strcpy(alertMessage,"Request successful!");
	}
	result=sceHttpReadData(httpRequest,response,*responseLength);
	if(result<0) {
		strcpy(alertMessage,"Error reading server data");
		sceHttpDeleteRequest(httpRequest);
		sceHttpDeleteConnection(httpConnection);
		return result;
	}
	response[result]=0;
	printf("Got %d bytes: '%.32s...\n",result,response);
	*responseLength=result;
	//printf("%s",httpResultBuf);
	result=sceHttpDeleteRequest(httpRequest);
	if(result<0) printf("Deleted request %08x\n",result);
	result=sceHttpDeleteConnection(httpConnection);
	if(result<0) printf("Deleted connection %08x\n",result);
	return *responseLength;		
}
#endif

int fetch(const char *url,char *response,int *responseLength)
{
	return post(url,0,0,response,responseLength);
}
